import unittest
from unittest import mock

from nepali.phone_number import (
    Operator,
    _get_area_code,
    _get_operator,
    _parse_landline_number,
    _parse_mobile_number,
    get_exact_number,
    is_landline_number,
    is_mobile_number,
    is_valid,
    parse,
)


class TestOperator(unittest.TestCase):
    def test_operator_with_invalid_data(self):
        with self.assertRaises(ValueError):
            Operator("Hello")

    def test_operator_with_valid_data(self):
        operator = Operator("Nepal Telecom")
        self.assertEqual(operator, Operator.NEPAL_TELECOM)

    def test_operator_str(self):
        operator = Operator.NEPAL_TELECOM
        self.assertEqual(str(operator), operator.value)

    def test_operator_repr(self):
        operator = Operator.NEPAL_TELECOM
        self.assertEqual(repr(operator), f"<Operator: {operator.value}>")


class TestPhoneNumberValidation(unittest.TestCase):
    def test_is_mobile_number(self):
        # True case
        self.assertEqual(is_mobile_number("9779842536789"), True)
        self.assertEqual(is_mobile_number("+9779842536789"), True)
        self.assertEqual(is_mobile_number("+977-9842536789"), True)
        self.assertEqual(is_mobile_number("9842536789"), True)
        self.assertEqual(is_mobile_number("9852536789"), True)
        self.assertEqual(is_mobile_number("9862536789"), True)
        self.assertEqual(is_mobile_number("9742536789"), True)
        self.assertEqual(is_mobile_number("9752536789"), True)
        self.assertEqual(is_mobile_number("9802536789"), True)
        self.assertEqual(is_mobile_number("9812536789"), True)
        self.assertEqual(is_mobile_number("9822536789"), True)
        self.assertEqual(is_mobile_number("9612536789"), True)
        self.assertEqual(is_mobile_number("9622536789"), True)
        self.assertEqual(is_mobile_number("9882536789"), True)
        self.assertEqual(is_mobile_number("9722536789"), True)
        self.assertEqual(is_mobile_number("9632536789"), True)

        # False case
        self.assertEqual(is_mobile_number(None), False)  # type: ignore
        self.assertEqual(is_mobile_number(""), False)
        self.assertEqual(is_mobile_number("test"), False)
        self.assertEqual(is_mobile_number("977test"), False)
        self.assertEqual(is_mobile_number("+977test"), False)
        self.assertEqual(is_mobile_number("+977-test"), False)
        self.assertEqual(is_mobile_number("9551088967"), False)
        self.assertEqual(is_mobile_number("+977-98425367899"), False)
        self.assertEqual(is_mobile_number("+977-984253678"), False)
        self.assertEqual(is_mobile_number("984253678"), False)
        self.assertEqual(is_mobile_number("98425367899"), False)

    def test_is_landline_number(self):
        # True case
        self.assertEqual(is_landline_number("14231481"), True)
        self.assertEqual(is_landline_number("014231481"), True)
        self.assertEqual(is_landline_number("0215154579"), True)
        self.assertEqual(is_landline_number("+977-142314819"), True)
        self.assertEqual(is_landline_number("0154225639"), True)
        self.assertEqual(is_landline_number("9770154225639"), True)
        self.assertEqual(is_landline_number("0565114679"), True)
        self.assertEqual(is_landline_number("021420451"), True)
        self.assertEqual(is_landline_number("21420451"), True)
        self.assertEqual(is_landline_number("+977-0565114679"), True)

        # False case
        self.assertEqual(is_landline_number(None), False)  # type: ignore
        self.assertEqual(is_landline_number(""), False)
        self.assertEqual(is_landline_number("test"), False)
        self.assertEqual(is_landline_number("977test"), False)
        self.assertEqual(is_landline_number("+977test"), False)
        self.assertEqual(is_landline_number("+977-test"), False)

    @mock.patch("nepali.phone_number.is_landline_number", return_value=False)
    @mock.patch("nepali.phone_number.is_mobile_number", return_value=False)
    def test_if__is_valid__returns_false_for_invalid_numbers(self, *_):
        self.assertEqual(is_valid("who_cares"), False)

    @mock.patch("nepali.phone_number.is_landline_number", return_value=True)
    @mock.patch("nepali.phone_number.is_mobile_number", return_value=False)
    def test_if__is_valid__returns_true_for_landline(self, *_):
        self.assertEqual(is_valid("who_cares"), True)

    @mock.patch("nepali.phone_number.is_landline_number", return_value=False)
    @mock.patch("nepali.phone_number.is_mobile_number", return_value=True)
    def test_if__is_valid__returns_true_for_mobile(self, *_):
        self.assertEqual(is_valid("who_cares"), True)

    @mock.patch("nepali.phone_number.is_landline_number", return_value=True)
    @mock.patch("nepali.phone_number.is_mobile_number", return_value=True)
    def test_if__is_valid__returns_true_if_both_mobile_and_landline_returns_true(
        self, *_
    ):
        self.assertEqual(is_valid("who_cares"), True)


class TestPhoneNumberFunctions(unittest.TestCase):
    def test_get_exact_number(self):
        self.assertEqual(get_exact_number("9842536789"), "9842536789")
        self.assertEqual(get_exact_number("9779842536789"), "9842536789")
        self.assertEqual(get_exact_number("+9779842536789"), "9842536789")
        self.assertEqual(get_exact_number("+977-9842536789"), "9842536789")
        self.assertEqual(get_exact_number(""), "")
        self.assertEqual(get_exact_number("test"), "test")
        self.assertEqual(get_exact_number("977test"), "test")
        self.assertEqual(get_exact_number("+977test"), "test")
        self.assertEqual(get_exact_number("+977-test"), "test")
        self.assertEqual(get_exact_number("+977-142314819"), "142314819")
        self.assertEqual(get_exact_number("0154225639"), "0154225639")
        self.assertEqual(get_exact_number("9770154225639"), "0154225639")
        self.assertEqual(get_exact_number("0565114679"), "0565114679")
        self.assertEqual(get_exact_number("984-2536789"), "9842536789")
        self.assertEqual(get_exact_number("984-2536-789"), "9842536789")
        with self.assertRaises(AttributeError):
            get_exact_number(None)  # type: ignore


class TestMobileNumberParser(unittest.TestCase):
    def test__get_operator__with_invalid_input(self):
        self.assertEqual(_get_operator("Hello there"), None)

    def test__get_operator__with_not_matching_input(self):
        self.assertEqual(_get_operator("9132536789"), None)

    def test__get_operator__for_ntc(self):
        self.assertEqual(_get_operator("9842536789"), Operator.NEPAL_TELECOM)
        self.assertEqual(_get_operator("9852536789"), Operator.NEPAL_TELECOM)
        self.assertEqual(_get_operator("9862536789"), Operator.NEPAL_TELECOM)
        # For CDMA (Nepal Telecom)
        self.assertEqual(_get_operator("9742536789"), Operator.NEPAL_TELECOM)
        self.assertEqual(_get_operator("9752536789"), Operator.NEPAL_TELECOM)

    def test__get_operator__for_ncell(self):
        self.assertEqual(_get_operator("9802536789"), Operator.NCELL)
        self.assertEqual(_get_operator("9812536789"), Operator.NCELL)
        self.assertEqual(_get_operator("9822536789"), Operator.NCELL)

    def test__get_operator__for_smart_cell(self):
        self.assertEqual(_get_operator("9612536789"), Operator.SMART_CELL)
        self.assertEqual(_get_operator("9622536789"), Operator.SMART_CELL)
        self.assertEqual(_get_operator("9882536789"), Operator.SMART_CELL)

    def test__get_operator__for_utl(self):
        self.assertEqual(_get_operator("9722536789"), Operator.UTL)

    def test__get_operator__for_hello_mobile(self):
        self.assertEqual(_get_operator("9632536789"), Operator.HELLO_MOBILE)

    @mock.patch("nepali.phone_number._get_operator", return_value="test")
    def test__parse_mobile_number_returns_valid_data(self, *_):
        data = _parse_mobile_number("+977-9842536789")
        self.assertEqual(data and data["type"], "Mobile")
        self.assertEqual(data and data["operator"], "test")
        self.assertEqual(data and data["number"], "9842536789")

    @mock.patch("nepali.phone_number._get_operator", return_value=None)
    def test__parse_mobile_number_on_invalid_operator(self, *_):
        data = _parse_mobile_number("+977-14231481")
        self.assertEqual(data, None)


class TestLandlineNumberParser(unittest.TestCase):
    def test__get_area_code_for_kathmandu(self):
        self.assertEqual(_get_area_code("0142314819"), "01")
        self.assertEqual(_get_area_code("0154225639"), "01")

    def test__get_area_code_for_biratnagar(self):
        self.assertEqual(_get_area_code("021420431"), "021")

    def test__get_area_code_for_chitwan(self):
        self.assertEqual(_get_area_code("056520198"), "056")

    def test__get_area_code_for_jhapa(self):
        self.assertEqual(_get_area_code("023455125"), "023")

    def test__get_area_code_for_rupandehi(self):
        self.assertEqual(_get_area_code("071540029"), "071")

    def test__get_area_code_for_kaski(self):
        self.assertEqual(_get_area_code("061463076"), "061")

    def test__get_area_code_for_dhading(self):
        self.assertEqual(_get_area_code("010520133"), "010")

    def test__get_area_code_for_syangja(self):
        self.assertEqual(_get_area_code("063420211"), "063")

    @mock.patch("nepali.phone_number._get_area_code", return_value="test")
    def test__parse_landline_number_returns_valid_data(self, *_):
        data = _parse_landline_number("0142314819")
        self.assertEqual(data["type"], "Landline")
        self.assertEqual(data["area_code"], "test")
        self.assertEqual(data["number"], "0142314819")

    @mock.patch("nepali.phone_number._get_area_code", return_value="test")
    def test__parse_landline_number_returns_valid_data_for_977(self, *_):
        data = _parse_landline_number("+977142314819")
        self.assertEqual(data["type"], "Landline")
        self.assertEqual(data["area_code"], "test")
        self.assertEqual(data["number"], "0142314819")


class TestParser(unittest.TestCase):
    def test_none_number_returns_none(self):
        self.assertEqual(parse(None), None)  # type: ignore

    def test_empty_number_returns_none(self):
        self.assertEqual(parse(""), None)

    @mock.patch("nepali.phone_number.is_mobile_number", return_value=True)
    @mock.patch("nepali.phone_number._parse_mobile_number", return_value="Mobile")
    def test_mobile_number_parsing(self, *_):
        self.assertEqual(parse("number"), "Mobile")

    @mock.patch("nepali.phone_number.is_mobile_number", return_value=False)
    @mock.patch("nepali.phone_number.is_landline_number", return_value=True)
    @mock.patch("nepali.phone_number._parse_landline_number", return_value="Landline")
    def test_landline_number_parsing(self, *_):
        self.assertEqual(parse("number"), "Landline")

    @mock.patch("nepali.phone_number.is_mobile_number", return_value=False)
    @mock.patch("nepali.phone_number.is_landline_number", return_value=False)
    def test_invalid_number_returns_none(self, *_):
        self.assertEqual(parse("number"), None)


class TestParseIntegration(unittest.TestCase):
    def test_parse_for_mobile_number(self):
        data = parse("+9779842536789")
        self.assertEqual(data and data["type"], "Mobile")
        self.assertEqual(data and data["number"], "9842536789")
        self.assertEqual(data and data["operator"], Operator.NEPAL_TELECOM)

    def test_parse_for_landline_number(self):
        data = parse("+977-142314819")
        self.assertEqual(data and data["type"], "Landline")
        self.assertEqual(data and data["number"], "0142314819")
        self.assertEqual(data and data["area_code"], "01")



from nepali.datetime.utils import to_nepalidate, to_nepalidatetime
from nepali.timezone import to_nepali_timezone, to_utc_timezone

__all__ = [
    "to_nepalidate",
    "to_nepalidatetime",
    "to_utc_timezone",
    "to_nepali_timezone",
]



from ._nepalinumber import nepalinumber
from ._number import NepaliNumber
from .utils import (
    add_comma,
    add_comma_english,
    convert_and_add_comma,
    english_to_nepali,
    nepali_to_english,
)

__all__ = [
    "NepaliNumber",
    "add_comma",
    "add_comma_english",
    "convert_and_add_comma",
    "english_to_nepali",
    "nepali_to_english",
    "nepalinumber",
]



"""
This module contains django templatetags for nepali date and time.
"""
import datetime
import warnings
from typing import Optional, Union

from django import template
from django.utils import timezone

from nepali.datetime import nepalidate as _nepalidate
from nepali.datetime import nepalidatetime as _nepalidatetime
from nepali.datetime import nepalihumanize as humanize
from nepali.exceptions import InvalidNepaliDateTimeObjectException
from nepali.utils import to_nepalidatetime

_DEFAULT_DATE_FORMAT = "%B %d, %Y, %A"
DEPRECIATION_WARNING_MESSAGE = (
    "The templatetag 'nepalidatetime' has been depreciated "
    "and will be removed in the future release. "
    "Please use `django-nepali` package."
)
_datetime = Union[datetime.date, datetime.datetime, _nepalidate, _nepalidatetime]
register = template.Library()


@register.filter(name="nepalidate")
def nepalidate(
    datetime_obj: _datetime, format: str = _DEFAULT_DATE_FORMAT
) -> Union[str, None]:
    """
    Renders the datetime object into nepali datetime format in 'en-US' locale (English).

    Usage:
    ```
    {{ datetime_obj|nepalidate }}
    {{ datetime_obj|nepalidate:"%Y-%m-%d" }}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Nepali datetime format in 'en-US' locale
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return nepalidate_en(datetime_obj, format=format)


@register.filter(name="nepalidate_en")
def nepalidate_en(
    datetime_obj: _datetime, format: str = _DEFAULT_DATE_FORMAT
) -> Union[str, None]:
    """
    Renders the datetime object into nepali datetime format in 'en-US' locale (English).

    Usage:
    ```
    {{ datetime_obj|nepalidate_en }}
    {{ datetime_obj|nepalidate_en:"%Y-%m-%d" }}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Nepali datetime format in 'en-US' locale
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    try:
        nepali_datetime_obj = to_nepalidatetime(datetime_obj)
        return nepali_datetime_obj.strftime_en(format)
    except InvalidNepaliDateTimeObjectException:
        return None


@register.filter(name="nepalidate_ne")
def nepalidate_ne(
    datetime_obj: _datetime, format: str = _DEFAULT_DATE_FORMAT
) -> Union[str, None]:
    """
    Renders the datetime object into nepali datetime format in 'ne' locale (Nepali).

    Usage:
    ```
    {{ datetime_obj|nepalidate_ne }}
    {{ datetime_obj|nepalidate_ne:"%Y-%m-%d" }}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Nepali datetime format in 'ne' locale
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    try:
        nepali_datetime_obj = to_nepalidatetime(datetime_obj)
        return nepali_datetime_obj.strftime_ne(format)
    except InvalidNepaliDateTimeObjectException:
        return None


@register.filter(name="nepalihumanize")
def nepalihumanize(
    datetime_obj: _datetime,
    threshold: Optional[int] = None,
    format: Optional[str] = None,
) -> Union[str, None]:
    """
    Renders the datetime object to a human readable form for 'ne' locale (Nepali).

    Usage:
    ```
    {{ datetime_obj|nepalihumanize }}
    ```

    :param datetime_obj: Datetime object
    :param threshold: Threshold in seconds that determines when to render
        the datetime object in the standard datetime format, optional
    :param format: Output format if threshold exceeded, optional
    :returns: Datetime object in human readable form
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    try:
        nepali_datetime_obj = to_nepalidatetime(datetime_obj)
        return humanize(nepali_datetime_obj, threshold=threshold, format=format)
    except InvalidNepaliDateTimeObjectException:
        return None


@register.simple_tag
def nepalinow(format: str = _DEFAULT_DATE_FORMAT) -> str:
    """
    Renders the current nepali datetime in 'en-US' locale (English).

    Usage:
    ```
    {% nepalinow %}
    {% nepalinow '%Y-%m-%d' %}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Current nepali datetime
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return nepalinow_en(format)


@register.simple_tag
def nepalinow_en(format: str = _DEFAULT_DATE_FORMAT) -> str:
    """
    Renders the current nepali datetime in 'en-US' locale (English).

    Usage:
    ```
    {% nepalinow_en %}
    {% nepalinow_en '%Y-%m-%d' %}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Current nepali datetime
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return to_nepalidatetime(timezone.now()).strftime(format)


@register.simple_tag
def nepalinow_ne(format: str = _DEFAULT_DATE_FORMAT) -> str:
    """
    Renders the current nepali datetime in 'ne' locale (Nepali).

    Usage:
    ```
    {% nepalinow_ne %}
    {% nepalinow_ne '%Y-%m-%d' %}
    ```

    :param datetime_obj: Datetime object
    :param format: Output format, defaults to "%B %d, %Y, %A"
    :returns: Current nepali datetime
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return to_nepalidatetime(timezone.now()).strftime_ne(format)



import datetime
import unittest

from nepali.datetime import nepalidate, nepalidatetime, nepalitime
from nepali.datetime.utils import to_nepalidate, to_nepalidatetime
from nepali.exceptions import InvalidNepaliDateTimeObjectException
from nepali.timezone import NepaliTimeZone


class TestNepaliDateTime(unittest.TestCase):
    #
    # nepalidate

    def test_nepalidate_now_and_today(self):
        today = nepalidate.today()
        now = nepalidate.now()
        self.assertEqual(today, now, msg="nepalidate today is not equals to now.")

    def test_nepalidate_comparison(self):
        nepalidate_obj1 = nepalidate(year=2051, month=4, day=28)

        # test equal date
        nepalidate_obj2 = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(
            nepalidate_obj1, nepalidate_obj2, "Same nepali date are not checked equals."
        )

        # test different date
        nepalidate_obj3 = nepalidate(year=2051, month=4, day=29)
        self.assertNotEqual(
            nepalidate_obj1,
            nepalidate_obj3,
            "Different nepali date are checked as equal.",
        )

        #
        # test addition

        # test with timedelta
        nt = nepalidate(year=2051, month=4, day=29)
        nt = nt + datetime.timedelta(days=1)
        self.assertEqual((nt.year, nt.month, nt.day), (2051, 4, 30))

        # test without timedelta
        with self.assertRaises(TypeError):
            nt = nt + 1

        #
        # test subtraction

        # test with nepali date
        td = nepalidate(year=2051, month=4, day=30) - nepalidate(
            year=2051, month=4, day=29
        )
        self.assertEqual(td.days, 1)

        # test with python date
        td = nepalidate(year=2051, month=4, day=30) - datetime.date(
            year=1994, month=8, day=13
        )
        self.assertEqual(td.days, 1)

        # test with timedelta
        nt = nepalidate(year=2051, month=4, day=30) - datetime.timedelta(days=1)
        self.assertEqual((nt.year, nt.month, nt.day), (2051, 4, 29))

        # test without timedelta
        with self.assertRaises(TypeError):
            nt = nt - 1

        #
        # test less than
        self.assertEqual(nepalidate(2051, 4, 29) < nepalidate(2051, 4, 30), True)
        self.assertEqual(nepalidate(2051, 4, 29) < nepalidate(2051, 4, 29), False)
        self.assertEqual(nepalidate(2051, 4, 29) < datetime.date(1994, 8, 14), True)
        self.assertEqual(nepalidate(2051, 4, 29) < datetime.date(1994, 8, 13), False)
        with self.assertRaises(TypeError):
            _ = nepalidate(2051, 4, 29) < 0

        #
        # test less than equal
        self.assertEqual(nepalidate(2051, 4, 29) <= nepalidate(2051, 4, 30), True)
        self.assertEqual(nepalidate(2051, 4, 29) <= nepalidate(2051, 4, 29), True)
        self.assertEqual(nepalidate(2051, 4, 30) <= nepalidate(2051, 4, 29), False)
        self.assertEqual(nepalidate(2051, 4, 29) <= datetime.date(1994, 8, 14), True)
        self.assertEqual(nepalidate(2051, 4, 29) <= datetime.date(1994, 8, 13), True)
        self.assertEqual(nepalidate(2051, 4, 30) <= datetime.date(1994, 8, 13), False)
        with self.assertRaises(TypeError):
            _ = nepalidate(2051, 4, 29) <= 0

        #
        # test greater than
        self.assertEqual(nepalidate(2051, 4, 30) > nepalidate(2051, 4, 29), True)
        self.assertEqual(nepalidate(2051, 4, 30) > nepalidate(2051, 4, 30), False)
        self.assertEqual(nepalidate(2051, 4, 30) > datetime.date(1994, 8, 14), False)
        self.assertEqual(nepalidate(2051, 4, 30) > datetime.date(1994, 8, 13), True)
        with self.assertRaises(TypeError):
            _ = nepalidate(2051, 4, 29) > 0

        #
        # test greater than equal
        self.assertEqual(nepalidate(2051, 4, 30) >= nepalidate(2051, 4, 30), True)
        self.assertEqual(nepalidate(2051, 4, 30) >= nepalidate(2051, 4, 29), True)
        self.assertEqual(nepalidate(2051, 4, 29) >= nepalidate(2051, 4, 30), False)
        self.assertEqual(nepalidate(2051, 4, 30) >= datetime.date(1994, 8, 14), True)
        self.assertEqual(nepalidate(2051, 4, 30) >= datetime.date(1994, 8, 13), True)
        self.assertEqual(nepalidate(2051, 4, 29) >= datetime.date(1994, 8, 14), False)
        with self.assertRaises(TypeError):
            _ = nepalidate(2051, 4, 29) >= 0

    def test_nepalidate_year_month_day(self):
        nepalidate_obj = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(nepalidate_obj.year, 2051)
        self.assertEqual(nepalidate_obj.month, 4)
        self.assertEqual(nepalidate_obj.day, 28)
        self.assertEqual(nepalidate_obj.weekday(), 5)

    def test_nepalidate_from_date(self):
        python_date_obj = datetime.date(year=1994, month=8, day=12)
        nepalidate_obj = nepalidate.from_date(python_date_obj)
        self.assertEqual(
            python_date_obj,
            nepalidate_obj.to_date(),
            "nepalidate from_date and to_date are not equal.",
        )

    def test_nepalidate_from_datetime(self):
        python_datetime_obj = datetime.datetime(
            year=1994, month=8, day=12, hour=4, minute=8, second=12
        )
        nepalidate_obj = nepalidate.from_datetime(python_datetime_obj)
        self.assertEqual(
            python_datetime_obj.date(),
            nepalidate_obj.to_date(),
            "nepalidate from_datetime and to_datetime are not equal.",
        )

    def test_nepalidate_strftime(self):
        nepalidate_obj = nepalidate(year=2051, month=4, day=28)
        self.assertEqual(nepalidate_obj.strftime("%Y-%m-%d"), "2051-04-28")
        self.assertEqual(
            nepalidate_obj.strftime("%a %A %w %d %b %B %m %y %Y"),
            "Fri Friday 5 28 Sharwan Sharwan 04 51 2051",
        )
        self.assertEqual(nepalidate_obj.strftime_en("%Y-%m-%d"), "2051-04-28")
        self.assertEqual(
            nepalidate_obj.strftime_en("%a %A %w %d %b %B %m %y %Y"),
            "Fri Friday 5 28 Sharwan Sharwan 04 51 2051",
        )
        self.assertEqual(nepalidate_obj.strftime_ne("%Y-%m-%d"), "२०५१-०४-२८")
        self.assertEqual(
            nepalidate_obj.strftime_ne("%a %A %w %d %b %B %m %y %Y"),
            "शुक्र शुक्रबार ५ २८ साउन साउन ०४ ५१ २०५१",
        )

    def test_nepalidate_strptime(self):
        nepalidatetime_obj = nepalidate.strptime("2078-01-12", format="%Y-%m-%d")
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

    #
    # nepalitime

    def test_nepalitime_now(self):
        nepalitime_obj, dt_now = nepalitime.now(), datetime.datetime.now()
        self.assertEqual(nepalitime_obj.hour, dt_now.hour)
        self.assertEqual(nepalitime_obj.minute, dt_now.minute)
        self.assertEqual(nepalitime_obj.second, dt_now.second)

    #
    # nepalidatetime

    def test_nepalidatetime_from_date(self):
        python_datetime_obj = datetime.datetime(
            year=1994, month=8, day=12, hour=5, minute=28, second=23
        )
        nepalidatetime_obj = nepalidatetime.from_date(python_datetime_obj)
        self.assertEqual(python_datetime_obj.date(), nepalidatetime_obj.to_date())

    def test_nepalidatetime_from_datetime(self):
        python_datetime_obj = datetime.datetime(
            year=1994,
            month=8,
            day=12,
            hour=5,
            minute=28,
            second=23,
            tzinfo=NepaliTimeZone(),
        )
        nepalidatetime_obj = nepalidatetime.from_datetime(python_datetime_obj)
        self.assertEqual(
            python_datetime_obj,
            nepalidatetime_obj.to_datetime(),
            msg="{} and {} are not equal.".format(
                python_datetime_obj, nepalidatetime_obj.to_datetime()
            ),
        )

    def test_nepalidatetime_strftime(self):
        nepalidatetime_obj = nepalidatetime(
            year=2051, month=4, day=28, hour=5, minute=28, second=23
        )
        self.assertEqual(
            nepalidatetime_obj.strftime("%Y-%m-%d %H:%M:%S"), "2051-04-28 05:28:23"
        )
        self.assertEqual(
            nepalidatetime_obj.strftime("%a %A %w %d %b %B %m %y %Y %H %I %p %M %S"),
            "Fri Friday 5 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23",
        )
        self.assertEqual(
            nepalidatetime_obj.strftime_en("%Y-%m-%d %H:%M:%S"), "2051-04-28 05:28:23"
        )
        self.assertEqual(
            nepalidatetime_obj.strftime_en("%a %A %w %d %b %B %m %y %Y %H %I %p %M %S"),
            "Fri Friday 5 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23",
        )
        self.assertEqual(
            nepalidatetime_obj.strftime_en(
                "%a %A %w %d %-d %b %B %m %-m %y %Y %H %-H %I %-I %p %M %-M %S %-S %%"
            ),
            "Fri Friday 5 28 28 Sharwan Sharwan 04 4 51 2051 05 5 05 5 AM 28 28 23 23 %",
        )
        self.assertEqual(
            nepalidatetime_obj.strftime_ne("%Y-%m-%d %H:%M:%S"), "२०५१-०४-२८ ०५:२८:२३"
        )
        self.assertEqual(
            nepalidatetime_obj.strftime_ne("%a %A %w %d %b %B %m %y %Y %H %I %p %M %S"),
            "शुक्र शुक्रबार ५ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३",
        )

    def test_nepalidatetime_strptime(self):
        nepalidatetime_obj = nepalidatetime.strptime(
            "2078-01-12 13:12", format="%Y-%m-%d %H:%M"
        )
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 12)

    def test_nepalidatetime_timedelta(self):
        nepalidatetime_obj1 = nepalidatetime(
            year=2051, month=4, day=28, hour=5, minute=28, second=23
        )
        nepalidatetime_obj2 = nepalidatetime(
            year=2051, month=4, day=29, hour=5, minute=28, second=23
        )
        nepalidatetime_obj1 = nepalidatetime_obj1 + datetime.timedelta(days=1)
        self.assertEqual(
            nepalidatetime_obj1,
            nepalidatetime_obj2,
            msg=f"{nepalidatetime_obj1} and {nepalidatetime_obj2} are not equal",
        )


class TestDatetimeUtils(unittest.TestCase):
    # to_nepalidate
    def test_to_nepalidate_raises_exception_on_invalid_input(self):
        with self.assertRaises(InvalidNepaliDateTimeObjectException):
            to_nepalidate("Invalid input")

    def test_to_nepalidate_from_python_date(self):
        np_date = to_nepalidate(datetime.date(2023, 2, 26))
        self.assertSequenceEqual(
            (np_date.year, np_date.month, np_date.day), (2079, 11, 14)
        )

    def test_to_nepalidate_from_python_datetime(self):
        np_date = to_nepalidate(datetime.datetime(2023, 2, 26, 1, 12, 13))
        self.assertSequenceEqual(
            (np_date.year, np_date.month, np_date.day), (2079, 11, 14)
        )

    def test_to_nepalidate_from_nepalidate(self):
        np_date = to_nepalidate(nepalidate(2079, 11, 14))
        self.assertSequenceEqual(
            (np_date.year, np_date.month, np_date.day), (2079, 11, 14)
        )

    def test_to_nepalidate_from_nepalidatetime(self):
        np_date = to_nepalidate(nepalidatetime(2079, 11, 14, 1, 12, 13))
        self.assertSequenceEqual(
            (np_date.year, np_date.month, np_date.day), (2079, 11, 14)
        )

    # to_nepalidatetime
    def test_to_nepalidatetime_raises_exception_on_invalid_input(self):
        with self.assertRaises(InvalidNepaliDateTimeObjectException):
            to_nepalidatetime("Invalid input")

    def test_to_nepalidatetime_from_python_date(self):
        np_date = to_nepalidatetime(datetime.date(2023, 2, 26))
        self.assertSequenceEqual(
            (
                np_date.year,
                np_date.month,
                np_date.day,
                np_date.hour,
                np_date.minute,
                np_date.second,
            ),
            (2079, 11, 14, 0, 0, 0),
        )

    def test_to_nepalidatetime_from_python_datetime(self):
        np_date = to_nepalidatetime(
            datetime.datetime(2023, 2, 26, 4, 30, 13, tzinfo=datetime.timezone.utc)
        )
        self.assertSequenceEqual(
            (
                np_date.year,
                np_date.month,
                np_date.day,
                np_date.hour,
                np_date.minute,
                np_date.second,
            ),
            (2079, 11, 14, 10, 15, 13),
        )

    def test_to_nepalidatetime_from_nepalidate(self):
        np_date = to_nepalidatetime(nepalidate(2079, 11, 14))
        self.assertSequenceEqual(
            (
                np_date.year,
                np_date.month,
                np_date.day,
                np_date.hour,
                np_date.minute,
                np_date.second,
            ),
            (2079, 11, 14, 0, 0, 0),
        )

    def test_to_nepalidatetime_from_nepalidatetime(self):
        np_date = to_nepalidatetime(nepalidatetime(2079, 11, 14, 1, 12, 13))
        self.assertSequenceEqual(
            (
                np_date.year,
                np_date.month,
                np_date.day,
                np_date.hour,
                np_date.minute,
                np_date.second,
            ),
            (2079, 11, 14, 1, 12, 13),
        )



import unittest

from nepali.char import nepali_to_english_text
from nepali.datetime import nepalidatetime
from nepali.datetime import parser as nepalidatetime_parser
from nepali.datetime.parser.validators import NepaliTimeRE, extract, transform, validate
from nepali.exceptions import FormatNotMatchException, InvalidDateTimeFormatException


class TestNepaliDateTimeParserValidators(unittest.TestCase):
    """
    Tests nepali date
    """

    def setUp(self) -> None:
        self.nepali_time_re = NepaliTimeRE()
        return super().setUp()

    # test extract

    def test_simple_extract(self):
        extracted_data = extract("2078-01-12", format="%Y-%m-%d")
        self.assertEqual(extracted_data.get("Y"), "2078")
        self.assertEqual(extracted_data.get("m"), "01")
        self.assertEqual(extracted_data.get("d"), "12")

    def test_complex_extract(self):
        extracted_data = extract(
            "Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23",
            format="%a %A %w %d %b %B %m %y %Y %H %I %p %M %S",
        )
        self.assertEqual(extracted_data.get("a"), "Wed")
        self.assertEqual(extracted_data.get("A"), "Wednesday")
        self.assertEqual(extracted_data.get("w"), "3")
        self.assertEqual(extracted_data.get("d"), "28")
        self.assertEqual(extracted_data.get("b"), "Sharwan")
        self.assertEqual(extracted_data.get("B"), "Sharwan")
        self.assertEqual(extracted_data.get("m"), "04")
        self.assertEqual(extracted_data.get("y"), "51")
        self.assertEqual(extracted_data.get("Y"), "2051")
        self.assertEqual(extracted_data.get("H"), "05")
        self.assertEqual(extracted_data.get("I"), "05")
        self.assertEqual(extracted_data.get("p"), "AM")
        self.assertEqual(extracted_data.get("M"), "28")
        self.assertEqual(extracted_data.get("S"), "23")

    def test_nepali_to_english_text_conversion(self):
        self.assertEqual(
            nepali_to_english_text(
                "बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३"
            ),
            "Wed Wednesday 3 28 Sharwan Sharwan 04 51 2051 05 05 AM 28 23",
        )

    def test_complex_extract_in_nepali(self):
        extracted_data = extract(
            "बुध बुधबार ३ २८ साउन साउन ०४ ५१ २०५१ ०५ ०५ शुभप्रभात २८ २३",
            format="%a %A %w %d %b %B %m %y %Y %H %I %p %M %S",
        )
        self.assertEqual(extracted_data.get("a"), "Wed")
        self.assertEqual(extracted_data.get("A"), "Wednesday")
        self.assertEqual(extracted_data.get("w"), "3")
        self.assertEqual(extracted_data.get("d"), "28")
        self.assertEqual(extracted_data.get("b"), "Sharwan")
        self.assertEqual(extracted_data.get("B"), "Sharwan")
        self.assertEqual(extracted_data.get("m"), "04")
        self.assertEqual(extracted_data.get("y"), "51")
        self.assertEqual(extracted_data.get("Y"), "2051")
        self.assertEqual(extracted_data.get("H"), "05")
        self.assertEqual(extracted_data.get("I"), "05")
        self.assertEqual(extracted_data.get("p"), "AM")
        self.assertEqual(extracted_data.get("M"), "28")
        self.assertEqual(extracted_data.get("S"), "23")

    # test transform
    def test_transform(self):
        data = transform(
            {
                "Y": "2078",
                "b": "MAngsir",
                "d": "12",
                "p": "pm",
                "I": "05",
                "M": "25",
                "S": "31",
                "f": "455",
                "a": "Wed",
            }
        )
        self.assertEqual(data.get("year"), 2078)
        self.assertEqual(data.get("month"), 8)
        self.assertEqual(data.get("day"), 12)
        self.assertEqual(data.get("hour"), 17)
        self.assertEqual(data.get("minute"), 25)
        self.assertEqual(data.get("second"), 31)
        self.assertEqual(data.get("microsecond"), 455000)
        self.assertEqual(data.get("weekday"), 3)

    #
    # test validate

    def test_validate(self):
        nepalidatetime_obj = validate("2078-01-12", format="%Y-%m-%d")
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

        nepalidatetime_obj = validate("2079-03-32", format="%Y-%m-%d")
        self.assertEqual(nepalidatetime_obj.year, 2079)
        self.assertEqual(nepalidatetime_obj.month, 3)
        self.assertEqual(nepalidatetime_obj.day, 32)

    def test_test_validate(self):
        nepalidatetime_obj = validate(
            "29 Jestha, 2078, 1:30 PM", format="%d %B, %Y, %I:%M %p"
        )
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 2)
        self.assertEqual(nepalidatetime_obj.day, 29)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 30)
        self.assertEqual(nepalidatetime_obj.second, 0)

    def test_validate_with_invalid_data(self):
        self.assertEqual(validate("2078-01-12", format="%k-%m-%d"), None)
        self.assertEqual(validate("2078-01-12", format="%m-%M-%Y"), None)


class TestNepaliDateTimeParser(unittest.TestCase):
    """
    Tests nepali datetime parser.
    """

    # test strptime
    def test_strptime(self):
        nepalidatetime_obj = nepalidatetime_parser.strptime(
            "2078-01-12", format="%Y-%m-%d"
        )
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 1)
        self.assertEqual(nepalidatetime_obj.day, 12)

    def test_strptime_fail(self):
        with self.assertRaises(FormatNotMatchException):
            nepalidatetime_parser.strptime("2078-01-12", format="%Y %d")

    # test parser

    def test_normal_string_parse(self):
        parsed_datetime = nepalidatetime_parser.parse("2071-01-24")
        test_datetime = nepalidatetime(2071, 1, 24)
        self.assertEqual(parsed_datetime, test_datetime)

    def test_complex_string_parse(self):
        nepalidatetime_obj = nepalidatetime_parser.parse("29 Jestha, 2078, 1:30 PM")
        self.assertEqual(nepalidatetime_obj.year, 2078)
        self.assertEqual(nepalidatetime_obj.month, 2)
        self.assertEqual(nepalidatetime_obj.day, 29)
        self.assertEqual(nepalidatetime_obj.hour, 13)
        self.assertEqual(nepalidatetime_obj.minute, 30)

    def test_parse_failed(self):
        with self.assertRaises(InvalidDateTimeFormatException):
            nepalidatetime_parser.parse("")



from nepali import number
from nepali.timezone import now

from .utils import to_nepalidatetime


class HumanizeDateTime:
    """
    HumanizeDate converts NepaliDateTime to nepali human readable form
    """

    __past_text = "अघि"
    __future_text = "पछि"
    __now_text = "भर्खरै"
    __year_text = "वर्ष"
    __month_text = "महिना"
    __day_text = "दिन"
    __hour_text = "घण्टा"
    __minute_text = "मिनेट"
    __second_text = "सेकेन्ड"

    def __init__(self, datetime_obj, **kwargs):
        """
        initializes humanize class
        datetime_obj: python datetime object to be humanized
        threshold (kwargs): threshold to be humanize
        format (kwargs): format to display behind threshold
        """
        self.datetime_obj = to_nepalidatetime(datetime_obj)

        self.threshold = kwargs.get("threshold")
        self.format = kwargs.get("format")

        # seconds after from now to datetime_obj
        self.seconds = 0

    def __calc_seconds(self):
        """calculates total seconds from now"""
        current_date_time = now()

        # TODO (@aj3sh): support datetime - nepalidatetime
        self.seconds = int(
            (current_date_time - self.datetime_obj.to_datetime()).total_seconds()
        )

        self.interval_tense = self.__past_text
        if self.seconds < 0:
            self.seconds *= -1
            self.interval_tense = self.__future_text

        return self.seconds

    def to_str(self):
        """returns humanize string"""
        seconds = self.__calc_seconds()  # calculating seconds

        if self.threshold is not None and seconds >= self.threshold:
            return self._get_datetime_str().strip()

        return self._get_humanize_str().strip()

    def _get_humanize_str(self):
        """
        returns humanize datetime
        """
        interval_value = 0
        interval_text = ""
        if self.seconds == 0:
            # now
            return self.__now_text

        elif self.seconds < 60:
            # seconds
            interval_value = self.seconds
            interval_text = self.__second_text

        elif self.seconds < 3600:
            # minute
            interval_value = self.seconds // 60
            interval_text = self.__minute_text

        elif self.seconds < 86400:
            # hour
            interval_value = self.seconds // 3600
            interval_text = self.__hour_text

        elif self.seconds < 2764800:
            # day
            interval_value = self.seconds // 86400
            interval_text = self.__day_text

        elif self.seconds < 31622400:
            # month
            interval_value = self.seconds // 2764800
            interval_text = self.__month_text

        else:
            # year
            interval_value = self.seconds // 31622400
            interval_text = self.__year_text

        interval_value = number.english_to_nepali(interval_value)
        return (
            str(interval_value) + " " + str(interval_text) + " " + self.interval_tense
        )

    def _get_datetime_str(self):
        """
        returns date in nepali characters
        """
        if not self.format:
            self.format = "%B %d, %Y"
        return self.datetime_obj.strftime_ne(self.format)

    def __str__(self):
        return self.to_str()

    def __repr__(self):
        return str(self)


def nepalihumanize(datetime_obj, threshold=None, format=None):
    """returns to humanize nepalidatetime"""
    return HumanizeDateTime(datetime_obj, threshold=threshold, format=format).to_str()



"""
Contains the class for the nepalinumber feature
"""

from typing import Any, Tuple, Type, Union

from .utils import NP_NUMBERS, NP_NUMBERS_SET, english_to_nepali


class nepalinumber:
    """
    Represents the nepali(devanagari) numbers and
    the features related to arithmetic operations
    on them
    """

    def __init__(self, value: Any) -> None:
        """
        Constructor/Initializer
        """
        self.__value = self.__parse(value)

    def get_parse_exception(
        self, obj: object, ex_class: Type[Exception] = ValueError
    ) -> Exception:
        """
        Returns the exception object to be raised when the parse is failed.
        The methods also sets a proper message to the exception class.

        :param obj: Object that is failed during the parse
        :type obj: object
        :param ex_class: Exception class type to be returned, defaults to ValueError
        :type ex_class: Type[Exception], optional
        :return: Exception object to be raised
        :rtype: Exception
        """
        return ex_class(
            f"could not convert {obj.__class__.__name__} to {self.__class__.__name__}: '{obj}'"
        )

    def __parse(self, value: Any) -> Union[int, float]:
        """
        Parses nepali number input into a valid value.

        Eg:
        >>> self.__parse("१२")
        12
        >>> self.__parse("१२.३")
        12.3
        >>> self.__parse(1)
        1
        >>> self.__parse("invalid")
        ValueError: could not convert str to nepalinumber: 'invalid'

        :param value: Value to be parsed.
        :return: returns value int or float
        :raises ValueError: If the value is invalid
        :raises TypeError: If the value object can't be parsed
        """
        if isinstance(value, int):
            return int(value)

        elif isinstance(value, float):
            return float(value)

        elif isinstance(value, str):
            return self.__parse_str(value)

        return self.__parse_object(value)

    def __parse_str(self, value: str) -> Union[int, float]:
        """
        Parses str object into int and float.
        This is a low level implementation.

        :raises ValueError: If the value is invalid
        """
        result: float = 0
        sign = 1
        decimal_found = False
        decimal_place: float = 1
        i = 0

        # for negative sign
        if value[0] == "-":
            sign = -1
            i = 1

        while i < len(value):
            # decimal number found
            if value[i] == ".":
                if decimal_found:
                    # decimal was already found
                    raise self.get_parse_exception(value) from None
                decimal_found = True
                i += 1
                continue

            digit = ord(value[i]) - ord("0")
            if digit < 0 or digit > 9:
                # checking nepali character
                if value[i] not in NP_NUMBERS_SET:
                    raise self.get_parse_exception(value) from None
                digit = NP_NUMBERS.index(value[i])

            if decimal_found:
                decimal_place /= 10
                result += digit * decimal_place
            else:
                result = result * 10 + digit

            i += 1
        return sign * result

    def __parse_object(self, obj: Any) -> Union[int, float]:
        """
        Parses object using __int__, __float__, and __str__.

        :raises TypeError: If the value object can't be parsed
        """
        try:
            if hasattr(obj, "__float__"):
                return float(obj)
            elif hasattr(obj, "__int__"):
                return int(obj)

            return self.__parse_str(str(obj))
        except (ValueError, TypeError):
            # object conversion must raise TypeError if fails
            raise self.get_parse_exception(obj, ex_class=TypeError) from None

    def __convert_or_return(self, obj) -> Union["nepalinumber", object]:
        """
        Will try to parse the given object and convert to nepalinumber
        else will return the same object

        :param obj: The object to convert

        :returns: Either a nepalinumber or the same object unchanged
        """
        try:
            return nepalinumber(obj)
        except (TypeError, ValueError):
            return obj

    def __str__(self) -> str:
        """
        Called when the object is called with functions
        like print or logger.debug()
        """
        return str(self.__value)

    def __repr__(self) -> str:
        return str(self.__value)

    def __int__(self) -> int:
        """
        Called when the object is typecasted into integer
        """
        return int(self.__value)

    def __float__(self) -> float:
        """
        Called when the object is typecasted into float
        """
        return float(self.__value)

    def __add(self, other) -> Union[int, float]:
        """
        Adds the value in the object with the passed object

        :param other: The other number/object to be added to the object
        :raises TypeError: Raised when unsupported data types
            are added to the nepalinumber object
        :return: A new nepalinumber object with the added values
        """
        if isinstance(other, nepalinumber):
            return self.__value + other.value

        return self.__value + other

    def __mul(self, other) -> Union[int, float]:
        """
        Multiplies the value in the object with the passed object

        :param other: The other number/object to be added to the object
        :raises TypeError: Raised when unsupported data types
            are multiplied to the nepalinumber object
        :return: A new nepalinumber object with the multiplied values
        """
        if isinstance(other, nepalinumber):
            return self.__value * other.value

        return self.__value * other

    def __eq__(self, other) -> bool:
        """
        Checks if nepalinumber is equal to another object

        :param other: The other number/object which is to be checked for
            equality against nepalinumber
        :return: True if equal else False
        """
        if isinstance(other, nepalinumber):
            return self.__value == other.value

        return self.__value == other

    def __ne__(self, other) -> bool:
        """
        Checks if nepalinumber is not equal to another object

        :param other: The other number/object which is to be checked for
            equality against nepalinumber
        :return: True if not equal else False
        """
        if isinstance(other, nepalinumber):
            return self.__value != other.value

        return self.__value != other

    def __neg__(self) -> "nepalinumber":
        """
        Returns the negative value of the nepalinumber value
        """
        return nepalinumber((-1) * self.__value)

    def __add__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the addition operator +  is used after
        the nepalinumber object

        :param other: The other number/object that is to be
            added to the value onto the nepalinumber object
        :raises TypeError: Raised when unsupported data types
            are added to the nepalinumber object
        :return: Returns the added value as a nepalinumber
            object
        """
        try:
            return self.__convert_or_return(self.__add(other))
        except TypeError:
            return NotImplemented

    def __radd__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the addition operator + is used before
        the nepalinumber object

        :param other: The other number/object that is to be
            added to the value onto the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            added to unsupported data types
        :return: Returns the added value as a nepalinumber
            object
        """
        try:
            return self.__convert_or_return(self.__add(other))
        except TypeError:
            return NotImplemented

    def __sub__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the subtraction operator - is used after
        the nepalinumber object

        :param other: The other number/object that is to be
            subtracted from the value in the nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            subtracted from nepalinumber object
        :return: Returns the subtracted number as a nepalinumber object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(self.__value - other.value)

            return self.__convert_or_return(self.__value - other)
        except TypeError:
            return NotImplemented

    def __rsub__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the subtraction operator - is used before
        the nepalinumber object

        :param other: The other number/object that is to get
            subtracted by the value in the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            subtracted from  unsupported data types
        :return: Returns the subtracted number as a nepalinumber
             object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(other.value - self.__value)

            return self.__convert_or_return(other - self.__value)
        except TypeError:
            return NotImplemented

    def __mul__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the multiplication operator * is used after
        the nepalinumber object

        :param other: The other number/object that is to be
            multiplied to the value onto the nepalinumber object
        :raises TypeError: Raised when unsupported data types
            are multiplied to the nepalinumber object
        :return: Returns the multiplied value as a nepalinumber
            object
        """
        try:
            if isinstance(other, str):
                return self.__value * other  # type: ignore

            return self.__convert_or_return(self.__mul(other))
        except TypeError:
            return NotImplemented

    def __rmul__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the multiplication operator * is used before
        the nepalinumber object

        :param other: The other number/object that is to be
            multiplied to the value onto the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            multiplied to unsupported data types
        :return: Returns the multiplied value as a nepalinumber
            object
        """
        try:
            if isinstance(other, str):
                return other * self.__value  # type: ignore

            return self.__convert_or_return(self.__mul(other))
        except TypeError:
            return NotImplemented

    def __truediv__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the division operator / is used after
        the nepalinumber object

        :param other: The other number/object that is to divide
            the value in the nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            used to divide nepalinumber object
        :return: Returns the quotient number as a nepalinumber object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(self.__value / other.value)

            return self.__convert_or_return(self.__value / other)
        except TypeError:
            return NotImplemented

    def __rtruediv__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the division operator / is used before
        the nepalinumber object

        :param other: The other number/object that is to get
            divided by the value in the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            used to divide unsupported data types
        :return: Returns the quotient number as a nepalinumber
             object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(other.value / self.__value)

            return self.__convert_or_return(other / self.__value)
        except TypeError:
            return NotImplemented

    def __floordiv__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the floor/integer division operator // is used
        after the nepalinumber object

        :param other: The other number/object that is to divide
            the value in the nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            used to divide nepalinumber object
        :return: Returns the quotient number as a nepalinumber object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(self.__value // other.value)

            return self.__convert_or_return(self.__value // other)
        except TypeError:
            return NotImplemented

    def __rfloordiv__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the floor/integer division operator // is used
        before the nepalinumber object

        :param other: The other number/object that is to get
            divided by the value in the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            used to divide unsupported data types
        :return: Returns the quotient number as a nepalinumber
             object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(other.value // self.__value)

            return self.__convert_or_return(other // self.__value)
        except TypeError:
            return NotImplemented

    def __mod__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the modulo operator % is used after
        the nepalinumber object

        :param other: The other number/object that is to be
            perform modulo division from the value in the
            nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            modulo divided from nepalinumber object
        :return: Returns the remainder number as a nepalinumber object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(self.__value % other.value)

            return self.__convert_or_return(self.__value % other)
        except TypeError:
            return NotImplemented

    def __rmod__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the modulo operator % is used before
        the nepalinumber object

        :param other: The other number/object that is to get
            modulo divided by the value in the nepalinumber object
        :raises TypeError: Raised when nepalinumber object is
            used modulo divide unsupported data types
        :return: Returns the remainder number as a nepalinumber
             object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(other.value % self.__value)

            return self.__convert_or_return(other % self.__value)
        except TypeError:
            return NotImplemented

    def __divmod__(
        self, other
    ) -> Tuple[Union["nepalinumber", object], Union["nepalinumber", object]]:
        """
        Called when the built-in function divmod() is used
        with nepalinumber as the dividend and other as divisor

        :param other: The other number/object that is to be
            divisor for the value in the nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            used as divisor for nepalinumber object
        :return: Returns a tuple of quotient and remainder
        """
        try:
            if isinstance(other, nepalinumber):
                quotient, remainder = divmod(self.__value, other.value)

            quotient, remainder = divmod(self.__value, other)

            return self.__convert_or_return(quotient), self.__convert_or_return(
                remainder
            )
        except TypeError:
            return NotImplemented

    def __rdivmod__(
        self, other
    ) -> Tuple[Union["nepalinumber", object], Union["nepalinumber", object]]:
        """
        Called when the built-in function divmod() is used
        with nepalinumber as the divisor and other as dividend

        :param other: The other number/object that is to be
            dividend for the value in the nepalinumber object
        :raises TypeError: Raised when unsupported data types are
            used as dividend for nepalinumber object
        :return: Returns a tuple of quotient and remainder
        """
        try:
            if isinstance(other, nepalinumber):
                quotient, remainder = divmod(other.value, self.__value)

            quotient, remainder = divmod(other, self.__value)

            return self.__convert_or_return(quotient), self.__convert_or_return(
                remainder
            )
        except TypeError:
            return NotImplemented

    def __pow__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the power operator **  is used after
        the nepalinumber object

        :param other: The other number/object that is to be
            powered to the value onto the nepalinumber object
        :raises TypeError: Raised when unsupported data types
            are powered to the nepalinumber object
        :return: Returns the powered by value as a nepalinumber
            object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(self.__value**other.value)

            return self.__convert_or_return(self.__value**other)
        except TypeError:
            return NotImplemented

    def __rpow__(self, other) -> Union["nepalinumber", object]:
        """
        Called when the power operator ** is used before
        the nepalinumber object

        :param other: The other number/object that is to be
            powered by the value onto the nepalinumber object
        :raises TypeError: Raised when unsupported data types
            are powered by the nepalinumber object
        :return: Returns the powered by value as a nepalinumber
            object
        """
        try:
            if isinstance(other, nepalinumber):
                return self.__convert_or_return(other.value**self.__value)

            return self.__convert_or_return(other**self.__value)
        except TypeError:
            return NotImplemented

    def str_ne(self) -> str:
        """
        Returns nepali (devanagari) format for the number

        :return: Stringified Nepali number
        """
        if not hasattr(self, "__str_ne"):
            self.__str_ne = english_to_nepali(self.__value)
        return self.__str_ne

    @property
    def value(self):
        return self.__value



from enum import Enum
from typing import List


class Location:
    def __init__(self, name: str, name_nepali: str):
        self.__name = name
        self.__name_nepali = name_nepali

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @property
    def name(self):
        return self.__name

    @property
    def name_nepali(self):
        return self.__name_nepali


class Province(Location):
    def __init__(self, name: str, name_nepali: str):
        super().__init__(name, name_nepali)
        self.__districts = []
        self.__municipalities = []

    def _add_district(self, district: "District") -> None:
        """Do not use outside of the model, automatically called from the models."""
        self.__districts.append(district)

    def _add_municipality(self, municipality: "Municipality") -> None:
        """Do not use outside of the model, automatically called from the models."""
        self.__municipalities.append(municipality)

    @property
    def districts(self) -> List["District"]:
        return self.__districts

    @property
    def municipalities(self) -> List["Municipality"]:
        return self.__municipalities


class District(Location):
    def __init__(self, province: Province, name: str, name_nepali: str):
        super().__init__(name, name_nepali)
        self.__province = province
        self.__municipalities = []
        self.__province._add_district(self)

    def _add_municipality(self, district: "Municipality") -> None:
        """Do not use outside of the model, automatically called from the models."""
        self.__municipalities.append(district)

    @property
    def province(self) -> Province:
        return self.__province

    @property
    def municipalities(self) -> List["Municipality"]:
        return self.__municipalities


class MunicipalityType(Enum):
    METROPOLITAN = "Metropolitan City"
    SUB_METROPOLITAN = "Sub-Metropolitan City"
    MUNICIPALITY = "Municipality"
    RURAL_MUNICIPALITY = "Rural Municipality"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class Municipality(Location):
    def __init__(
        self,
        district: District,
        name: str,
        name_nepali: str,
        municipality_type: MunicipalityType,
    ):
        super().__init__(name, name_nepali)
        self.__district = district
        self.__municipality_type = municipality_type
        self.__district._add_municipality(self)
        self.__district.province._add_municipality(self)

    @property
    def province(self) -> Province:
        return self.district.province

    @property
    def district(self) -> District:
        return self.__district

    @property
    def municipality_type(self) -> MunicipalityType:
        return self.__municipality_type



"""
To run only this unit test use the command below.

    python -m unittest nepali/tests/test_date_converter.py -v
"""
import unittest

from nepali.date_converter import converter


class TestNepaliDateConverter(unittest.TestCase):
    # test date range
    def test_english_date_range(self):
        self.assertEqual(converter.en_min_year(), 1944)
        self.assertEqual(converter.en_max_year(), 2042)

    def test_nepali_date_range(self):
        self.assertEqual(converter.np_min_year(), 2000)
        self.assertEqual(converter.np_max_year(), 2099)

    # english_to_nepali
    def test_converter_english_to_nepali_raise_exception_on_max_year_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(2060, 1, 4)

    def test_converter_english_to_nepali_raise_exception_on_min_year_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(1920, 1, 4)

    def test_converter_english_to_nepali_raise_exception_on_min_month_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(2023, 0, 4)

    def test_converter_english_to_nepali_raise_exception_on_max_month_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(2023, 13, 4)

    def test_converter_english_to_nepali_raise_exception_on_min_day_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(2023, 1, 0)

    def test_converter_english_to_nepali_raise_exception_on_max_day_range(self):
        with self.assertRaises(Exception):
            converter.english_to_nepali(2023, 1, 40)

    def test_converter_english_to_nepali_returns_valid_past_nepali_date(self):
        y, m, d = converter.english_to_nepali(1994, 8, 13)
        self.assertEqual(y, 2051)
        self.assertEqual(m, 4)
        self.assertEqual(d, 29)

    def test_converter_english_to_nepali_returns_valid_recent_nepali_date(self):
        y, m, d = converter.english_to_nepali(2023, 1, 28)
        self.assertEqual(y, 2079)
        self.assertEqual(m, 10)
        self.assertEqual(d, 14)

    def test_converter_english_to_nepali_returns_valid_future_nepali_date(self):
        y, m, d = converter.english_to_nepali(2030, 11, 26)
        self.assertEqual(y, 2087)
        self.assertEqual(m, 8)
        self.assertEqual(d, 10)

    def test_converter_english_to_nepali_for_min_edge_date(self):
        y, m, d = converter.english_to_nepali(1944, 1, 1)
        self.assertEqual(y, 2000)
        self.assertEqual(m, 9)
        self.assertEqual(d, 17)

    def test_converter_english_to_nepali_for_max_edge_date(self):
        y, m, d = converter.english_to_nepali(2042, 12, 31)
        self.assertEqual(y, 2099)
        self.assertEqual(m, 9)
        self.assertEqual(d, 16)

    # nepali_to_english
    def test_converter_nepali_to_englishReturnErrorOnMaxYearRange(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(3000, 1, 4)

    def test_converter_nepali_to_english_raise_exception_on_min_year_range(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(1920, 1, 4)

    def test_converter_nepali_to_english_raise_exception_on_min_month_range(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(2079, 0, 4)

    def test_converter_nepali_to_english_raise_exception_on_max_month_range(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(2079, 13, 4)

    def test_converter_nepali_to_english_raise_exception_on_min_day_range(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(2079, 1, 0)

    def test_converter_nepali_to_english_raise_exception_on_max_day_range(self):
        with self.assertRaises(Exception):
            converter.nepali_to_english(2079, 1, 40)

    def test_converter_nepali_to_englishReturnsValidPastEnglishDate(self):
        y, m, d = converter.nepali_to_english(2051, 4, 29)
        self.assertEqual(y, 1994)
        self.assertEqual(m, 8)
        self.assertEqual(d, 13)

    def test_converter_nepali_to_englishReturnsValidRecentEnglishDate(self):
        y, m, d = converter.nepali_to_english(2079, 10, 14)
        self.assertEqual(y, 2023)
        self.assertEqual(m, 1)
        self.assertEqual(d, 28)

    def test_converter_nepali_to_englishReturnsValidFutureEnglishDate(self):
        y, m, d = converter.nepali_to_english(2087, 8, 10)
        self.assertEqual(y, 2030)
        self.assertEqual(m, 11)
        self.assertEqual(d, 26)

    def test_converter_nepali_to_englishReturnsValidEnglishLeapYearDate(self):
        y, m, d = converter.nepali_to_english(2080, 12, 15)
        self.assertEqual(y, 2024)
        self.assertEqual(m, 3)
        self.assertEqual(d, 28)

    def test_converter_nepali_to_english_for_min_edge_date(self):
        y, m, d = converter.nepali_to_english(2000, 1, 1)
        self.assertEqual(y, 1943)
        self.assertEqual(m, 4)
        self.assertEqual(d, 14)

    def test_converter_nepali_to_english_for_max_edge_date(self):
        y, m, d = converter.nepali_to_english(2099, 12, 30)
        self.assertEqual(y, 2043)
        self.assertEqual(m, 4)
        self.assertEqual(d, 13)



"""
To run only this unit test use the command below.

    python -m unittest nepali/tests/test_nepaliweek.py -v
"""
import unittest

from nepali.datetime import nepaliweek


class TestNepaliWeek(unittest.TestCase):
    # parse
    def test_nepaliweek_parses_int(self):
        self.assertEqual(nepaliweek(0).value, 0)
        self.assertEqual(nepaliweek(6).value, 6)

    def test_nepaliweek_fails_for_int_less_than_0(self):
        with self.assertRaises(ValueError) as ex:
            nepaliweek(-100)
        self.assertEqual(str(ex.exception), "Invalid week: -100")

    def test_nepaliweek_fails_for_int_more_than_6(self):
        with self.assertRaises(ValueError) as ex:
            nepaliweek(7)
        self.assertEqual(str(ex.exception), "Invalid week: 7")
        with self.assertRaises(ValueError):
            nepaliweek(100)

    def test_nepaliweek_parses_str_int_number(self):
        week = nepaliweek("1")
        self.assertEqual(week.value, 1)

    def test_nepaliweek_fails_for_str_int_out_of_range(self):
        with self.assertRaises(ValueError):
            nepaliweek("-100")
        with self.assertRaises(ValueError):
            nepaliweek("7")
        with self.assertRaises(ValueError):
            nepaliweek("100")

    def test_nepaliweek_fails_for_invalid_str(self):
        with self.assertRaises(ValueError):
            nepaliweek("hello")
        with self.assertRaises(ValueError):
            nepaliweek("1.0")

    def test_nepaliweek_parse_valid_week_in_english(self):
        self.assertEqual(nepaliweek("Sunday").value, 0)
        self.assertEqual(nepaliweek("Monday").value, 1)
        self.assertEqual(nepaliweek("Tuesday").value, 2)
        self.assertEqual(nepaliweek("Wednesday").value, 3)
        self.assertEqual(nepaliweek("Thursday").value, 4)
        self.assertEqual(nepaliweek("Friday").value, 5)
        self.assertEqual(nepaliweek("Saturday").value, 6)

    def test_nepaliweek_parse_valid_week_in_english_lower_case(self):
        self.assertEqual(nepaliweek("sunday").value, 0)
        self.assertEqual(nepaliweek("monday").value, 1)
        self.assertEqual(nepaliweek("tuesday").value, 2)
        self.assertEqual(nepaliweek("wednesday").value, 3)
        self.assertEqual(nepaliweek("thursday").value, 4)
        self.assertEqual(nepaliweek("friday").value, 5)
        self.assertEqual(nepaliweek("saturday").value, 6)

    def test_nepaliweek_parse_valid_week_in_english_case_insensitive(self):
        self.assertEqual(nepaliweek("sUnday").value, 0)
        self.assertEqual(nepaliweek("moNday").value, 1)
        self.assertEqual(nepaliweek("tueSday").value, 2)
        self.assertEqual(nepaliweek("wednEsday").value, 3)
        self.assertEqual(nepaliweek("thursDay").value, 4)
        self.assertEqual(nepaliweek("fridaY").value, 5)
        self.assertEqual(nepaliweek("saTurday").value, 6)

    def test_nepaliweek_parse_valid_week_in_english_abbreviated_name(self):
        self.assertEqual(nepaliweek("Sun").value, 0)
        self.assertEqual(nepaliweek("Mon").value, 1)
        self.assertEqual(nepaliweek("Tue").value, 2)
        self.assertEqual(nepaliweek("Wed").value, 3)
        self.assertEqual(nepaliweek("Thu").value, 4)
        self.assertEqual(nepaliweek("Fri").value, 5)
        self.assertEqual(nepaliweek("Sat").value, 6)

    def test_nepaliweek_parse_valid_week_in_nepali(self):
        self.assertEqual(nepaliweek("आइतबार").value, 0)
        self.assertEqual(nepaliweek("सोमबार").value, 1)
        self.assertEqual(nepaliweek("मंगलबार").value, 2)
        self.assertEqual(nepaliweek("बुधबार").value, 3)
        self.assertEqual(nepaliweek("बिहीबार").value, 4)
        self.assertEqual(nepaliweek("शुक्रबार").value, 5)
        self.assertEqual(nepaliweek("शनिबार").value, 6)

    def test_nepaliweek_parse_valid_week_in_nepali_abbreviated_name(self):
        self.assertEqual(nepaliweek("आइत").value, 0)
        self.assertEqual(nepaliweek("सोम").value, 1)
        self.assertEqual(nepaliweek("मंगल").value, 2)
        self.assertEqual(nepaliweek("बुध").value, 3)
        self.assertEqual(nepaliweek("बिही").value, 4)
        self.assertEqual(nepaliweek("शुक्र").value, 5)
        self.assertEqual(nepaliweek("शनि").value, 6)

    def test_nepaliweek_fails_for_invalid_data_type(self):
        with self.assertRaises(ValueError):
            nepaliweek(1.0)  # type: ignore
        with self.assertRaises(ValueError):
            nepaliweek(2j + 1)  # type: ignore
        with self.assertRaises(ValueError):
            nepaliweek([])  # type: ignore

    def test_nepaliweek_str(self):
        self.assertEqual(str(nepaliweek(1)), nepaliweek(1).name)

    def test_nepaliweek_repr(self):
        self.assertEqual(repr(nepaliweek(1)), "<nepaliweek: 1>")

    def test_nepaliweek_name(self):
        self.assertEqual(nepaliweek(0).name, "Sunday")
        self.assertEqual(nepaliweek(6).name, "Saturday")

    def test_nepaliweek_abbr(self):
        self.assertEqual(nepaliweek(0).abbr, "Sun")
        self.assertEqual(nepaliweek(6).abbr, "Sat")

    def test_nepaliweek_name_ne(self):
        self.assertEqual(nepaliweek(0).name_ne, "आइतबार")
        self.assertEqual(nepaliweek(6).name_ne, "शनिबार")

    def test_nepaliweek_abbr_ne(self):
        self.assertEqual(nepaliweek(0).abbr_ne, "आइत")
        self.assertEqual(nepaliweek(6).abbr_ne, "शनि")

    def test_nepaliweek_variable_cache(self):
        self.assertEqual(id(nepaliweek(0)), id(nepaliweek(0)))
        self.assertEqual(id(nepaliweek(6)), id(nepaliweek(6)))

    def test_nepaliweek_int(self):
        self.assertEqual(int(nepaliweek(0)), 0)
        self.assertEqual(int(nepaliweek(6)), 6)

    def test_nepaliweek_equal_with_nepaliweek(self):
        self.assertEqual(nepaliweek(0), nepaliweek(0))
        self.assertEqual(nepaliweek(6), nepaliweek(6))

    def test_nepaliweek_not_equal(self):
        self.assertNotEqual(nepaliweek(0), nepaliweek(2))
        self.assertNotEqual(nepaliweek(1), 2)

    def test_nepaliweek_equal_with_int(self):
        self.assertEqual(nepaliweek(0), 0)
        self.assertEqual(nepaliweek(6), 6)
        self.assertEqual(1, nepaliweek(1))
        self.assertEqual(6, nepaliweek(6))

    def test_nepaliweek_not_equal_with_invalid_object(self):
        self.assertNotEqual(nepaliweek(1), "invalid")



import unittest

from nepali import locations
from nepali.locations.models import (
    District,
    Location,
    Municipality,
    MunicipalityType,
    Province,
)
from nepali.locations.utils import (
    _filter_location,
    get_district,
    get_municipality,
    get_province,
)


class TestLocations(unittest.TestCase):
    def test_locations_count(self):
        self.assertEqual(len(locations.provinces), 7)
        self.assertEqual(len(locations.districts), 77)
        self.assertEqual(len(locations.municipalities), 753)

    def test_utils_filter_location(self):
        # checking if no argument is passed
        with self.assertRaises(ValueError):
            _filter_location(locations=locations.provinces)

        # checking pattern match
        province = _filter_location(locations=locations.provinces, name="Bagmati")
        self.assertEqual(province.name, "Bagmati Province")

        # checking ignore-case match
        province = _filter_location(locations=locations.provinces, name="bagmati")
        self.assertEqual(province.name, "Bagmati Province")

        # checking with invalid name
        province = _filter_location(locations=locations.provinces, name="invalid")
        self.assertEqual(province, None)

        # checking using exact True with invalid match
        province = _filter_location(
            locations=locations.provinces, name="Bagmati", exact=True
        )
        self.assertEqual(province, None)

        # checking using exact True with a valid match
        province = _filter_location(
            locations=locations.provinces, name="Bagmati Province", exact=True
        )
        self.assertEqual(province.name, "Bagmati Province")

        # checking assertion if exact and multiple is passes together
        with self.assertRaises(ValueError):
            _filter_location(
                locations=locations.provinces,
                name="Bagmati Province",
                exact=True,
                multiple=True,
            )

        # checking multiple value match
        provinces = _filter_location(
            locations=locations.provinces, name="Province", multiple=True
        )
        self.assertEqual(type(provinces), list)
        self.assertEqual(len(provinces), 7)

        # checking multiple value with invalid match
        provinces = _filter_location(
            locations=locations.provinces, name="invalid", multiple=True
        )
        self.assertEqual(len(provinces), 0)

        # checking pattern match with nepali name
        province = _filter_location(locations=locations.provinces, name_nepali="बागमती")
        self.assertEqual(province.name_nepali, "बागमती प्रदेश")

        # checking using exact True with a valid nepali match
        province = _filter_location(
            locations=locations.provinces, name_nepali="बागमती प्रदेश", exact=True
        )
        self.assertEqual(province.name_nepali, "बागमती प्रदेश")


class TestLocationsUtils(unittest.TestCase):
    def test_utils_get_province(self):
        province = get_province(name="Bagmati")
        self.assertEqual(province.name, "Bagmati Province")

    def test_utils_get_district(self):
        district = get_district(name="Kathmandu")
        self.assertEqual(district.name, "Kathmandu")

    def test_utils_get_municipality(self):
        municipality = get_municipality(name="Kathmandu")
        self.assertEqual(municipality.name, "Kathmandu Metropolitan City")


class TestLocationsModels(unittest.TestCase):
    def test_location_data(self):
        location = Location("test_en", "test_ne")
        self.assertEqual(location.name, "test_en")
        self.assertEqual(location.name_nepali, "test_ne")

    def test_location_str(self):
        location = Location("test_en", "test_ne")
        self.assertEqual(str(location), "test_en")

    def test_location_repr(self):
        location = Location("test_en", "test_ne")
        self.assertEqual(repr(location), "test_en")

    def test_districts_from_province(self):
        province = Province("Province1", "Province1")
        district = District(province, "District1", "District1")
        self.assertListEqual(province.districts, [district])

    def test_municipalities_from_province_and_district(self):
        province = Province("Province1", "Province1")
        district = District(province, "District1", "District1")
        municipality = Municipality(
            district, "Municipality1", "Municipality1", MunicipalityType.METROPOLITAN
        )
        self.assertListEqual(province.municipalities, [municipality])
        self.assertListEqual(district.municipalities, [municipality])

    def test_municipality_province_district_and_type(self):
        province = Province("Province1", "Province1")
        district = District(province, "District1", "District1")
        municipality = Municipality(
            district, "Municipality1", "Municipality1", MunicipalityType.METROPOLITAN
        )
        self.assertEqual(municipality.province, province)
        self.assertEqual(municipality.district, district)
        self.assertEqual(municipality.municipality_type, MunicipalityType.METROPOLITAN)

    # MunicipalityType
    def test_municipality_type_with_invalid_data(self):
        with self.assertRaises(ValueError):
            MunicipalityType("Hello")

    def test_municipality_type_with_valid_data(self):
        municipality_type = MunicipalityType("Metropolitan City")
        self.assertEqual(municipality_type, MunicipalityType.METROPOLITAN)

    def test_municipality_type_str(self):
        municipality_type = MunicipalityType.RURAL_MUNICIPALITY
        self.assertEqual(str(municipality_type), municipality_type.value)

    def test_municipality_type_repr(self):
        municipality_type = MunicipalityType.SUB_METROPOLITAN
        self.assertEqual(repr(municipality_type), municipality_type.value)



import datetime
from typing import Union

from .constants import NEPAL_TIMEZONE


class NepaliTimeZone(datetime.tzinfo):
    """
    NepaliTimeZone: "Asia/Kathmandu", +05:45
    """

    def utcoffset(self, dt):
        return self.dst(dt) + datetime.timedelta(hours=5, minutes=45)

    def dst(self, dt):
        return datetime.timedelta(0)

    def tzname(self, dt):
        return NEPAL_TIMEZONE

    def __str__(self):
        return NEPAL_TIMEZONE

    def __repr__(self):
        return NEPAL_TIMEZONE

    def __eq__(self, o: object) -> bool:
        return isinstance(o, self.__class__)


def get_timezone() -> Union[datetime.tzinfo, None]:
    """
    Returns current device's timezone.
    Timezone of the machine.
    """
    return datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo


def now() -> datetime.datetime:
    """Returns current datetime object of the device"""
    return datetime.datetime.now(get_timezone())


def utc_now() -> datetime.datetime:
    """Returns UTC time datetime object"""
    return datetime.datetime.now(datetime.timezone.utc)


def to_utc_timezone(datetime_obj: datetime.datetime) -> datetime.datetime:
    """Changes the timezone of the given datetime object to UTC."""
    if type(datetime_obj) != datetime.datetime:
        # Not a datetime object
        return datetime_obj

    if not hasattr(datetime_obj, "tzinfo") or not datetime_obj.tzinfo:
        datetime_obj = datetime_obj.replace(tzinfo=get_timezone())
    return datetime_obj.astimezone(datetime.timezone.utc)


def to_nepali_timezone(datetime_obj: datetime.datetime) -> datetime.datetime:
    """Changes the timezone of the given datetime object to NepaliTimeZone."""
    if type(datetime_obj) != datetime.datetime:
        # Not a datetime object
        return datetime_obj

    if not hasattr(datetime_obj, "tzinfo") or not datetime_obj.tzinfo:
        datetime_obj = datetime_obj.replace(tzinfo=get_timezone())
    return datetime_obj.astimezone(NepaliTimeZone())



"""
This files contains constants for the py-Nepali project.
"""

# Nepali timezone
NEPAL_TIMEZONE = "Asia/Kathmandu"

# Nepali months in english
NEPALI_MONTHS_EN = (
    "Baishakh",
    "Jestha",
    "Ashad",
    "Sharwan",
    "Bhadra",
    "Ashwin",
    "Kartik",
    "Mangsir",
    "Poush",
    "Magh",
    "Falgun",
    "Chaitra",
)

# Nepali months in Nepali (devanagari)
NEPALI_MONTHS_NE = (
    "बैशाख",
    "जेठ",
    "असार",
    "साउन",
    "भदौ",
    "असोज",
    "कात्तिक",
    "मंसिर",
    "पुस",
    "माघ",
    "फागुन",
    "चैत",
)

# Week names in english
WEEKS_EN = (
    "Sunday",
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
)

# Week abbreviation name in english
WEEKS_ABBR_EN = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat")

# Weeks names in Nepali (devanagari)
WEEKS_NE = (
    "आइतबार",
    "सोमबार",
    "मंगलबार",
    "बुधबार",
    "बिहीबार",
    "शुक्रबार",
    "शनिबार",
)

# Week abbreviation name in Nepali (devanagari)
WEEKS_ABBR_NE = ("आइत", "सोम", "मंगल", "बुध", "बिही", "शुक्र", "शनि")



from typing import Any

NP_NUMBERS = ["०", "१", "२", "३", "४", "५", "६", "७", "८", "९"]
NP_NUMBERS_SET = set(NP_NUMBERS)


def english_to_nepali(number: Any) -> str:
    """
    Converts english number to nepali.
    """
    number = str(number)
    converted_number = []
    for n in number:
        num = ord(n) - ord("0")
        if num in range(0, 10):
            converted_number.append(NP_NUMBERS[num])
        else:
            converted_number.append(n)
    return "".join(converted_number)


def nepali_to_english(number: Any) -> str:
    """
    Converts nepali number to english.
    """
    number = str(number)
    converted_number = []
    for n in number:
        if n in NP_NUMBERS_SET:
            converted_number.append(str(NP_NUMBERS.index(n)))
        else:
            converted_number.append(n)
    return "".join(converted_number)


def add_comma_english(number: Any) -> str:
    """
    Adds comma in english style
    Eg. 123456789 => 123,456,789
    """
    return f"{int(number):,}"


def add_comma(number: Any, convert=False) -> str:
    """
    Adds comma in nepali style
    Eg. 123456789 => 12,34,56,789

    :param number Any: Number to be converted
    :param convert bool: If true converts english number to nepali
    """
    if convert:
        number = english_to_nepali(number)
    else:
        number = str(number)

    number_with_comma = []
    counter = 0
    for nepali_number_char in list(str(number))[::-1]:
        if counter == 3 or (counter != 1 and (counter - 1) % 2 == 0):
            number_with_comma.append(",")
        number_with_comma.append(nepali_number_char)
        counter += 1

    return "".join(number_with_comma[::-1])


def convert_and_add_comma(number: Any) -> str:
    """
    Converts the number into nepali text and adds comma to it.
    """
    return add_comma(number, convert=True)



import re
from functools import partial

from ._locations import districts, municipalities, provinces

__all__ = [
    "get_province",
    "get_district",
    "get_municipality",
]


def _filter_location(
    *, locations, name=None, name_nepali=None, exact=False, multiple=False
):
    if not name and not name_nepali:
        raise ValueError("name or name_nepali must be passed")

    if exact and multiple:
        raise ValueError("Both the exact and multiple cannot be true")

    # making name as lower case
    if name:
        name = name.lower()

    # taking field and value for filtering
    # eg. field = name and value ="Bagmati Province"
    field, value = ("name", name) if name else ("name_nepali", name_nepali)

    # filtering data
    if exact:
        filtered_locations = [
            location
            for location in locations
            if getattr(location, field).lower() == value
        ]
    else:
        pattern = re.compile(rf".*{value}.*")
        filtered_locations = [
            location
            for location in locations
            if re.match(pattern, getattr(location, field).lower())
        ]

    # returning data if not filtered location is available
    if len(filtered_locations) == 0 and not multiple:
        return None

    return filtered_locations if multiple else filtered_locations[0]


get_province = partial(_filter_location, locations=provinces)
get_district = partial(_filter_location, locations=districts)
get_municipality = partial(_filter_location, locations=municipalities)



import datetime
from typing import Any

from nepali.exceptions import InvalidNepaliDateTimeObjectException

from ._datetime import nepalidate, nepalidatetime


def to_nepalidatetime(datetime_object: Any) -> nepalidatetime:
    """
    Converts nepalidate, datetime.datetime, datetime.date to nepalidatetime.

    :param datetime_object: Object to be converted into nepalidatetime
    :return: nepalidatetime
    :raises InvalidNepaliDateTimeObjectException: If the input data is not a date time objects
    """
    if isinstance(datetime_object, nepalidatetime):
        return datetime_object
    elif isinstance(datetime_object, nepalidate):
        return nepalidatetime.from_nepali_date(datetime_object)
    elif isinstance(datetime_object, datetime.datetime):
        return nepalidatetime.from_datetime(datetime_object)
    elif isinstance(datetime_object, datetime.date):
        return nepalidatetime.from_date(datetime_object)
    raise InvalidNepaliDateTimeObjectException(
        "Argument must be instance of nepalidate or nepalidatetime or datetime.datetime or datetime.date"
    )


def to_nepalidate(datetime_object: Any) -> nepalidate:
    """
    Converts nepalidate, datetime.datetime, datetime.date to nepalidate.

    :param datetime_object: Object to be converted into nepalidate
    :return: nepalidate
    :raises InvalidNepaliDateTimeObjectException: If the input data is not a date time objects
    """
    return to_nepalidatetime(datetime_object).date()



import datetime as pythonDateTime

from nepali import number
from nepali.exceptions import (
    InvalidDateFormatException,
    InvalidNepaliDateTimeObjectException,
)

from ._datetime import nepalidate, nepalidatetime
from ._nepalimonth import nepalimonth
from ._nepaliweek import nepaliweek


class NepaliDateTimeFormatter:
    """
    NepaliDateTimeFormatter: formats nepali datetime to string ( using strftime )
    """

    # format according to python's datetime with class method
    format_map = {
        "a": "weekdayHalf",
        "A": "weekdayFull",
        "w": "weekdayNumber",
        "d": "day",
        "-d": "day_nonzero",
        "b": "monthFull",
        "B": "monthFull",
        "m": "monthNumber",
        "-m": "monthNumber_nonzero",
        "y": "yearHalf",
        "Y": "yearFull",
        "H": "hour24",
        "-H": "hour24_nonzero",
        "I": "hour12",
        "-I": "hour12_nonzero",
        "p": "ampm",
        "M": "minute",
        "-M": "minute_nonzero",
        "S": "second",
        "-S": "second_nonzero",
    }

    def __init__(self, datetime_object, devanagari=False):
        # TODO: Change variable npDateTime into snake case: `np_date_time`
        if type(datetime_object) == nepalidatetime:
            self.npDateTime = datetime_object
        elif type(datetime_object) == nepalidate:
            self.npDateTime = datetime_object.to_nepalidatetime()
        elif type(datetime_object) == pythonDateTime.date:
            self.npDateTime = nepalidatetime.from_date(datetime_object)
        elif type(datetime_object) == pythonDateTime.datetime:
            self.npDateTime = nepalidatetime.from_datetime(datetime_object)
        else:
            raise InvalidNepaliDateTimeObjectException(
                "Argument must be instance of nepalidate or nepalidatetime or datetime.datetime or datetime.date"
            )

        self.devanagari = devanagari

    def __str__(self):
        return str(self.npDateTime)

    def get_str(self, format):
        """generates formatted string"""
        i, n = 0, len(format)
        time_str = []
        try:
            while i < n:
                ch = format[i]
                i += 1
                if ch == "%":
                    if i < n:
                        ch = format[i]

                        if ch == "%":
                            # for % character
                            time_str.append("%")

                        elif ch == "-":
                            # special mid characters eg. "-" for non zero padded values
                            special_ch = ch
                            if i + 1 < n:
                                i += 1
                                ch = format[i]
                                time_str.append(
                                    getattr(self, self.get_format_map(special_ch + ch))
                                )
                        else:
                            # mapping % forwarded character
                            time_str.append(getattr(self, self.get_format_map(ch)))
                        i += 1
                else:
                    time_str.append(ch)
        except InvalidDateFormatException as e:
            raise e
        except Exception:
            raise Exception("Unable to convert NepaliDateTime to str")
        time_str = "".join(time_str)

        return time_str

    def get_format_map(self, ch: str) -> str:
        if ch not in self.format_map:
            raise InvalidDateFormatException(f"Invalid Date format %{ch}")
        return self.format_map[ch]

    @property
    def weekdayHalf(self):
        """
        %a
        """
        week = nepaliweek(self.npDateTime.weekday())
        if not self.devanagari:
            return week.abbr
        return week.abbr_ne

    @property
    def weekdayFull(self):
        """
        %A
        """
        week = nepaliweek(self.npDateTime.weekday())
        if not self.devanagari:
            return week.name
        return week.name_ne

    @property
    def weekdayNumber(self):
        """
        %w
        """
        if not self.devanagari:
            return str(self.npDateTime.weekday())
        return number.english_to_nepali(self.npDateTime.weekday())

    @property
    def day(self):
        """
        %d
        """
        day = str(self.npDateTime.day)
        if len(day) < 2:
            day = "0" + day
        if not self.devanagari:
            return str(day)
        return number.english_to_nepali(day)

    @property
    def day_nonzero(self):
        """
        %-D
        """
        day = str(self.npDateTime.day)
        if not self.devanagari:
            return str(day)
        return number.english_to_nepali(day)

    @property
    def monthFull(self):
        """
        %B or %b
        """
        month = nepalimonth(self.npDateTime.month)
        if not self.devanagari:
            return month.name
        return month.name_ne

    @property
    def monthNumber(self):
        """
        %m
        """
        month = str(self.npDateTime.month)
        if len(month) < 2:
            month = "0" + month
        if not self.devanagari:
            return str(month)
        return number.english_to_nepali(month)

    @property
    def monthNumber_nonzero(self):
        """
        %-m
        """
        month = str(self.npDateTime.month)
        if not self.devanagari:
            return str(month)
        return number.english_to_nepali(month)

    @property
    def yearHalf(self):
        """
        %y
        """
        if not self.devanagari:
            return str(self.npDateTime.year)[2:]
        return number.english_to_nepali(str(self.npDateTime.year)[2:])

    @property
    def yearFull(self):
        """
        %Y
        """
        if not self.devanagari:
            return str(self.npDateTime.year)
        return number.english_to_nepali(self.npDateTime.year)

    @property
    def hour24(self):
        """
        %H
        """
        hour = str(self.npDateTime.hour)
        if len(hour) < 2:
            hour = "0" + hour
        if not self.devanagari:
            return str(hour)
        return number.english_to_nepali(hour)

    @property
    def hour24_nonzero(self):
        """
        %-H
        """
        hour = self.npDateTime.hour
        if not self.devanagari:
            return str(hour)
        return number.english_to_nepali(hour)

    @property
    def hour12(self):
        """
        %I
        """
        hour = self.npDateTime.hour
        if hour > 12:
            hour = hour - 12
        if hour == 0:
            hour = 12
        hour = str(hour)
        if len(hour) < 2:
            hour = "0" + hour

        if not self.devanagari:
            return str(hour)
        return number.english_to_nepali(hour)

    @property
    def hour12_nonzero(self):
        """
        %-I
        """
        hour = self.npDateTime.hour
        if hour > 12:
            hour = hour - 12
        if hour == 0:
            hour = 12
        hour = str(hour)
        if not self.devanagari:
            return str(hour)
        return number.english_to_nepali(hour)

    @property
    def ampm(self):
        """
        %p
        """
        if not self.devanagari:
            ampm = "AM"
            if self.npDateTime.hour > 12:
                ampm = "PM"
            return str(ampm)

        ampm = ""
        if self.npDateTime.hour < 12:
            ampm = "शुभप्रभात"
        elif self.npDateTime.hour >= 12 and self.npDateTime.hour < 18:
            ampm = "मध्यान्ह"
        else:
            ampm = "अपरान्ह"
        return str(ampm)

    @property
    def minute(self):
        """
        %M
        """
        minute = str(self.npDateTime.minute)
        if len(minute) < 2:
            minute = "0" + minute
        if not self.devanagari:
            return str(minute)
        return number.english_to_nepali(minute)

    @property
    def minute_nonzero(self):
        """
        %-M
        """
        minute = str(self.npDateTime.minute)
        if not self.devanagari:
            return str(minute)
        return number.english_to_nepali(minute)

    @property
    def second(self):
        """
        %S
        """
        second = str(self.npDateTime.second)
        if len(second) < 2:
            second = "0" + second
        if not self.devanagari:
            return str(second)
        return number.english_to_nepali(second)

    @property
    def second_nonzero(self):
        """
        %-S
        """
        second = str(self.npDateTime.second)
        if not self.devanagari:
            return str(second)
        return number.english_to_nepali(second)



from nepali.exceptions import FormatNotMatchException, InvalidDateTimeFormatException

from .validators import validate

__all__ = [
    "strptime",
    "parse",
]

_standard_datetime_format_CACHE = None


def strptime(datetime_str, format):
    """
    Parses nepalidatetime str with the corresponding format.
    returns nepalidatetime object.
    """
    nepalidatetime_object = validate(datetime_str, format)
    if nepalidatetime_object is None:
        raise FormatNotMatchException(
            "Datetime string did not match with the given format."
        )
    return nepalidatetime_object


def _get_standard_formats():
    global _standard_datetime_format_CACHE
    if _standard_datetime_format_CACHE is not None:
        return _standard_datetime_format_CACHE

    STANDARD_DATE_FORMAT = [
        "%Y-%m-%d",
        "%d-%m-%Y",
        "%Y %m %d",
        "%Y %B %d",
        "%d %B %Y",
        "%d %B, %Y",
        "%B %d, %Y",
        "%Y/%m/%d",
        "%d/%m/%Y",
    ]

    STANDARD_TIME_FORMAT = [
        "%H:%M",
        "%I:%M %p",
        "%I %p",
        "%I%p",
        "%H:%M:%S",
        "%H:%M:%S:%f",
    ]

    _standard_datetime_format_CACHE = STANDARD_DATE_FORMAT.copy()
    for time_format in STANDARD_TIME_FORMAT:
        for date_format in STANDARD_DATE_FORMAT:
            _standard_datetime_format_CACHE += [
                f"{date_format} {time_format}",
                f"{date_format}, {time_format}",
            ]
    return _standard_datetime_format_CACHE


def parse(datetime_str):
    """
    parses nepali datetime
    eg. parse('2078-10-12') => <NepaliDateTime: 2078-10-12>
    """
    standard_formats = _get_standard_formats()
    for format in standard_formats:
        try:
            return strptime(datetime_str, format=format)
        except FormatNotMatchException:
            pass

    raise InvalidDateTimeFormatException("Invalid format to parse nepali datetime.")



from functools import cached_property
from typing import Any, Dict, Optional, Union

from nepali.constants import NEPALI_MONTHS_EN, NEPALI_MONTHS_NE


class NepaliMonthMeta(type):
    _cache: Dict[int, "nepalimonth"] = {}

    def __call__(cls, month: Union[int, str], *args, **kwargs) -> "nepalimonth":
        """
        Parses the month data and manages the cache.

        :param month: An integer or string representing the month.
        :type month: Union[int, str]
        :return: An instance of the nepalimonth class representing the given month.
        :rtype: nepalimonth
        :raises ValueError: If the given month is invalid.
        """
        value: Optional[int] = None
        value = None

        if isinstance(month, int):
            value = int(month)
        elif isinstance(month, str):
            try:
                value = cls._parse_str(month)
            except ValueError:
                pass

        # checking if month is valid
        if value is None or value < 1 or value > 12:
            raise ValueError(f"Invalid month: {month}")

        # checking cache
        if value not in cls._cache:
            cls._cache[value] = super(NepaliMonthMeta, cls).__call__(
                value, *args, **kwargs
            )

        return cls._cache[value]

    @staticmethod
    def _parse_str(month: str) -> int:
        """
        Parses str value of the month and returns int.

        :param month: A string representing the month.
        :type month: str
        :return: An integer representing the month.
        :rtype: int
        :raises ValueError: If the given string does not represent a valid month.
        """
        if month.isdigit():
            return int(month)

        month = month.capitalize()
        month_names = NEPALI_MONTHS_EN + NEPALI_MONTHS_NE
        try:
            index = month_names.index(month)
        except ValueError:
            raise ValueError(f"Invalid month name: {month}")

        return (index % 12) + 1


class nepalimonth(metaclass=NepaliMonthMeta):
    """
    Represents Nepali month: Baishakh, Jestha, ..., Chaitra.
    Baishakh: 1,
    Jestha: 2,
    ...
    Chaitra: 12

    >>> nepalimonth(1)
    >>> nepalimonth("Baishakh")
    >>> nepalimonth("बैशाख")

    :param Union[int, str] month: Month data to be parsed.

    :raises ValueError: The value is invalid.
    """

    def __init__(self, month) -> None:
        self.__value = month

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<nepalimonth: {self.__value}>"

    def __int__(self) -> int:
        return self.value

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, nepalimonth):
            return self.value == other.value
        elif isinstance(other, int):
            return self.value == other

        return False

    @cached_property
    def value(self) -> int:
        return self.__value

    @cached_property
    def name(self) -> str:
        """Month's english name"""
        return NEPALI_MONTHS_EN[self.__value - 1]

    @cached_property
    def name_ne(self) -> str:
        """Month's nepali name"""
        return NEPALI_MONTHS_NE[self.__value - 1]



import warnings

from .utils import add_comma, add_comma_english, english_to_nepali, nepali_to_english


# Backward compatibility support for legacy NepaliNumber
class NepaliNumber:
    @classmethod
    def convert_and_add_comma(cls, number):
        warnings.warn(
            message="NepaliNumber.convert_and_add_comma has been moved to "
            "`convert_and_add_comma. This function is depreciated and will "
            "be removed in the future release.",
            category=DeprecationWarning,
        )
        return add_comma(english_to_nepali(number))

    @staticmethod
    def convert(num):
        warnings.warn(
            message="NepaliNumber.convert has been moved to `english_to_nepali. "
            "This function is depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return english_to_nepali(num)

    @staticmethod
    def revert(num):
        warnings.warn(
            message="NepaliNumber.revert has been moved to `nepali_to_english. "
            "This function is depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepali_to_english(num)

    @staticmethod
    def add_comma(number):
        warnings.warn(
            message="NepaliNumber.add_comma has been moved to `add_comma. "
            "This function is depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return add_comma(number)

    @staticmethod
    def add_comma_english(number):
        warnings.warn(
            message="NepaliNumber.add_comma_english has been moved to "
            "`add_comma_english. This function is depreciated and will "
            "be removed in the future release.",
            category=DeprecationWarning,
        )
        return add_comma_english(number)



from typing import List, Tuple

from .models import District, Municipality, MunicipalityType, Province


def _loadData() -> Tuple[List[Province], List[District], List[Municipality]]:
    from ._data import _location_data

    provinces = []
    districts = []
    municipalities = []
    for province_data in _location_data:
        province = Province(
            name=province_data["name"], name_nepali=province_data["name_nepali"]
        )
        provinces.append(province)

        for district_data in province_data["districts"]:
            district = District(
                province=province,
                name=district_data["name"],
                name_nepali=district_data["name_nepali"],
            )
            districts.append(district)

            for municipality_data in district_data["municipalities"]:
                municipality = Municipality(
                    district=district,
                    name=municipality_data["name"],
                    name_nepali=municipality_data["name_nepali"],
                    municipality_type=MunicipalityType(
                        municipality_data["municipality_type"]
                    ),
                )
                municipalities.append(municipality)

    return provinces, districts, municipalities


provinces, districts, municipalities = _loadData()



import datetime
import unittest
from unittest.mock import patch

from nepali.datetime import HumanizeDateTime, nepalidate, nepalidatetime, nepalihumanize
from nepali.exceptions import InvalidNepaliDateTimeObjectException
from nepali.timezone import NepaliTimeZone

REF_TIME = datetime.datetime(2015, 1, 1, 10, 15, tzinfo=NepaliTimeZone())


@patch("nepali.datetime._humanize.now", return_value=REF_TIME)
class TestNepaliHumanizeFunction(unittest.TestCase):
    def test_nepalihumanize_for_just_now(self, *_):
        humanize_str = nepalihumanize(REF_TIME)
        self.assertEqual(humanize_str, "भर्खरै")

    def test_nepalihumanize_for_seconds(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(seconds=10))
        self.assertEqual(humanize_str, "१० सेकेन्ड अघि")

    def test_nepalihumanize_for_minutes(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(minutes=10))
        self.assertEqual(humanize_str, "१० मिनेट अघि")

    def test_nepalihumanize_for_hours(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(hours=10))
        self.assertEqual(humanize_str, "१० घण्टा अघि")

    def test_nepalihumanize_for_days(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(days=10))
        self.assertEqual(humanize_str, "१० दिन अघि")

    def test_nepalihumanize_for_months(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(days=32))
        self.assertEqual(humanize_str, "१ महिना अघि")

    def test_nepalihumanize_for_years(self, *_):
        humanize_str = nepalihumanize(REF_TIME - datetime.timedelta(days=366))
        self.assertEqual(humanize_str, "१ वर्ष अघि")

    def test_nepalihumanize_for_future(self, *_):
        humanize_str = nepalihumanize(REF_TIME + datetime.timedelta(days=10))
        self.assertEqual(humanize_str, "१० दिन पछि")

    def test_nepalihumanize_returns_date_after_threshold(self, *_):
        humanize_str = nepalihumanize(
            REF_TIME - datetime.timedelta(seconds=10), threshold=10
        )
        self.assertEqual(humanize_str, "पुस १७, २०७१")


class TestHumanizeDateTimeClass(unittest.TestCase):
    def test_nepali_humanize_raise_exception_on_invalid_input(self, *_):
        with self.assertRaises(InvalidNepaliDateTimeObjectException):
            HumanizeDateTime("test")

        with self.assertRaises(InvalidNepaliDateTimeObjectException):
            HumanizeDateTime(None)

        with self.assertRaises(InvalidNepaliDateTimeObjectException):
            HumanizeDateTime("2017-12-04")

    def test_nepali_humanize_supports_python_date(self, *_):
        humanize_obj = HumanizeDateTime(datetime.date.today())
        self.assertEqual(type(humanize_obj.to_str()), str)

    def test_nepali_humanize_supports_python_datetime(self, *_):
        humanize_obj = HumanizeDateTime(datetime.datetime.now())
        self.assertEqual(type(humanize_obj.to_str()), str)

    def test_nepali_humanize_supports_nepalidate(self, *_):
        humanize_obj = HumanizeDateTime(nepalidate.today())
        self.assertEqual(type(humanize_obj.to_str()), str)

    def test_nepali_humanize_supports_nepalidatetime(self, *_):
        humanize_obj = HumanizeDateTime(nepalidatetime.now())
        self.assertEqual(type(humanize_obj.to_str()), str)

    @patch("nepali.datetime._humanize.now", return_value=REF_TIME)
    def test_test_nepali_humanize_str_and_repr(self, *_):
        humanize_obj = HumanizeDateTime(REF_TIME + datetime.timedelta(days=10))
        self.assertEqual(str(humanize_obj), "१० दिन पछि")
        self.assertEqual(repr(humanize_obj), "१० दिन पछि")



"""
setup file for nepali package

- Building a package

    pip install build
    python -m build


- Publishing a package
You must have twine installed in your system. `pip install twine`

    python setup.py sdist bdist_wheel
    twine upload dist/*

"""
import os
import sys

import setuptools

GITHUB_URL = "https://github.com/opensource-nepal/py-nepali"
CHANGELOG_URL = "https://github.com/opensource-nepal/py-nepali/blob/main/CHANGELOG.md"

with open("README.md", "r") as fh:
    long_description = fh.read()


if sys.argv[-1] == "publish":
    if os.system("pip3 freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("rm -rf dist")
    os.system("python3 setup.py sdist bdist_wheel")
    os.system("twine upload dist/*")
    sys.exit()


setuptools.setup(
    name="nepali",
    version="1.1.0",
    license="MIT",
    author="opensource-nepal",
    author_email="aj3sshh@gmail.com, sugatbajracharya49@gmail.com",
    description="nepalidatetime compatible with python's datetime feature. "
    "Converting nepali date to english, parsing nepali datetime, "
    "nepali timezone, and timedelta support in nepali datetime",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=[
        "nepali date conversion",
        "convert date",
        "nepali date time",
        "python convert date",
        "parse nepali date time",
    ],
    url=GITHUB_URL,
    packages=setuptools.find_packages(exclude=["tests*"]),
    test_suite="nepali.tests",
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Source": GITHUB_URL,
        "Changelog": CHANGELOG_URL,
    },
)



"""
This module contains django templatetags for nepali number.
"""
import warnings
from typing import Any

from django import template

from nepali.number import (
    add_comma,
    add_comma_english,
    convert_and_add_comma,
    english_to_nepali,
)

register = template.Library()

DEPRECIATION_WARNING_MESSAGE = (
    "The templatetag 'nepalinumber' has been depreciated "
    "and will be removed in the future release. "
    "Please use `django-nepali` package."
)


@register.filter(name="nepalinumber")
def nepalinumber(value: Any) -> str:
    """
    Converts the number into nepali number and renders it.

    Usage:
    ```
    {{ number|nepalinumber }}
    ```

    :param value: Number to be converted
    :returns: Nepali output of given number
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return english_to_nepali(value)


@register.filter(name="nepalinumber_with_comma")
def nepalinumber_with_comma(value: Any) -> str:
    """
    Converts the number into nepali number and renders it.

    Usage:
    ```
    {{ number|nepalinumber_with_comma }}
    ```
    Basically same as `{{ number|nepalinumber|nepali_comma }}`

    :param value: Number to be converted and commas added
    :returns: Nepali output of given number with commas
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return convert_and_add_comma(value)


@register.filter(name="nepali_comma")
def nepali_comma(value: Any) -> str:
    """
    Renders the given value with commas added in Nepali style without converting the number.

    Usage:
    ```
    {{ number|nepali_comma }}
    ```

    :param value: Number to be added with commas
    :returns: Output of given number with commas
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return add_comma(value)


@register.filter(name="english_comma")
def english_comma(value: Any) -> str:
    """
    Renders the given value with commas added in English style without converting the number.

    Usage:
    ```
    {{ number|english_comma }}
    ```

    :param value: Number to be added with commas
    :returns: Output of given number with commas
    """
    warnings.warn(
        message=DEPRECIATION_WARNING_MESSAGE,
        category=DeprecationWarning,
    )
    return add_comma_english(value)



import warnings

from . import number
from .datetime._nepalimonth import nepalimonth
from .datetime._nepaliweek import nepaliweek

MONTHS_MAP = {
    "बैशाख": "Baishakh",
    "जेठ": "Jestha",
    "असार": "Ashad",
    "साउन": "Sharwan",
    "भदौ": "Bhadra",
    "असोज": "Ashwin",
    "कात्तिक": "Kartik",
    "मंसिर": "Mangsir",
    "पुस": "Poush",
    "माघ": "Magh",
    "फागुन": "Falgun",
    "चैत": "Chaitra",
}

DAYS_MAP = {
    "आइतबार": "Sunday",
    "सोमबार": "Monday",
    "मंगलबार": "Tuesday",
    "बुधबार": "Wednesday",
    "बिहीबार": "Thursday",
    "शुक्रबार": "Friday",
    "शनिबार": "Saturday",
}

HALF_DAYS_MAP = {
    "आइत": "Sun",
    "सोम": "Mon",
    "मंगल": "Tue",
    "बुध": "Wed",
    "बिही": "Thu",
    "शुक्र": "Fri",
    "शनि": "Sat",
}

AM_PM_MAP = {
    "शुभप्रभात": "AM",
    "मध्यान्ह": "PM",
    "अपरान्ह": "PM",
}


class NepaliChar:
    @staticmethod
    def number(num):
        warnings.warn(
            message="NepaliChar.number has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return number.english_to_nepali(num)

    @staticmethod
    def day(day):
        warnings.warn(
            message="NepaliChar.day has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepaliweek(day).name_ne

    @staticmethod
    def half_day(day):
        warnings.warn(
            message="NepaliChar.half_day has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepaliweek(day).abbr_ne

    @staticmethod
    def month(month):
        warnings.warn(
            message="NepaliChar.month has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepalimonth(month).name_ne


class EnglishChar:
    @staticmethod
    def day(day):
        warnings.warn(
            message="EnglishChar.day has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepaliweek(day).name

    @staticmethod
    def half_day(day):
        warnings.warn(
            message="EnglishChar.half_day has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepaliweek(day).abbr

    @staticmethod
    def month(month):
        warnings.warn(
            message="EnglishChar.month has been depreciated and will be removed in the future release.",
            category=DeprecationWarning,
        )
        return nepalimonth(month).name


def nepali_to_english_text(text):
    # TODO: optimization

    # replacing months
    for k, v in MONTHS_MAP.items():
        text = text.replace(k, v)

    # replacing days
    for k, v in DAYS_MAP.items():
        text = text.replace(k, v)

    # replacing half days
    for k, v in HALF_DAYS_MAP.items():
        text = text.replace(k, v)

    for k, v in AM_PM_MAP.items():
        text = text.replace(k, v)

    return number.nepali_to_english(text)



import re
from enum import Enum
from typing import Union

_mobile_number_re = re.compile(r"^(?:\+977|977)?(?:-)?(?:98|97|96)\d{8}$")
_landline_number_re = re.compile(
    r"^(?:\+977|977)?(?:-)?(?:0)?(?:[01][1-9]|2[13-9]|[3-9]\d)\d{6,7}$"
)


class Operator(Enum):
    NEPAL_TELECOM = "Nepal Telecom"
    NCELL = "Ncell"
    SMART_CELL = "Smart Cell"
    UTL = "UTL"
    HELLO_MOBILE = "Hello Mobile"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"<Operator: {self.value}>"


def is_mobile_number(number: str) -> bool:
    """
    Returns True is the input number is mobile number.

    >>> is_mobile = is_mobile_number(number)
    >>> if is_mobile:
    >>>     ...
    """
    try:
        return bool(_mobile_number_re.match(number))
    except Exception:
        return False


def is_landline_number(number: str) -> bool:
    """
    Returns True is the input number is mobile number.

    >>> is_mobile = is_mobile_number(number)
    >>> if is_mobile:
    >>>     ...
    """
    try:
        return bool(_landline_number_re.match(number))
    except Exception:
        return False


def is_valid(number: str) -> bool:
    return is_mobile_number(number) or is_landline_number(number)


def get_exact_number(number: str) -> str:
    # replacing start 977
    if number.startswith("977"):
        number = number.replace("977", "")
    # replacing +977 and all -
    return number.replace("+977", "").replace("-", "")


def parse(number: str):
    """
    Parse and returns the details of the phone number: `dict`.
    The return data may vary between mobile and landline numbers.
    If the number is invalid it returns `None`.

    If you want to make sure you get a valid response, please use
    `is_valid`, `is_mobile_number`, and `is_landline_number`.

    :return:
    {
        "type": "Mobile" | "Landline",
        "number": "XXXXXXX",
        ...
    }
    """
    if not number and type(number) != str:
        return None

    number = number.replace("-", "")

    # checking if mobile number
    if is_mobile_number(number):
        return _parse_mobile_number(number)

    if is_landline_number(number):
        return _parse_landline_number(number)

    return None


def _get_operator(number: str) -> Union[Operator, None]:
    """
    Returns operator from the number.
    NOTE: The number should be 10digit mobile number.
    """
    starting_number = number[:3]

    # NTC
    if starting_number in ["984", "985", "986", "974", "975"]:
        return Operator.NEPAL_TELECOM

    # NCELL
    if starting_number in ["980", "981", "982"]:
        return Operator.NCELL

    # Smart Cell
    if starting_number in ["961", "962", "988"]:
        return Operator.SMART_CELL

    # UTL
    if starting_number == "972":
        return Operator.UTL

    # Hello Mobile
    if starting_number == "963":
        return Operator.HELLO_MOBILE

    return None


def _parse_mobile_number(number: str):
    """
    Parse and returns mobile number details.
    :return:
    {
        "type": "Mobile",
        "number": "98XXXXXXXX",
        "operator": <Operator>
    }
    """
    number = get_exact_number(number)
    operator = _get_operator(number)

    if not operator:
        return None

    detail = {
        "type": "Mobile",
        "number": number,
        "operator": operator,
    }
    return detail


def _get_area_code(number) -> str:
    """
    Returns area code of the number.
    NOTE: The number should be landline number without +977/977.
    """
    code = number[:3]

    # Kathmandu, Lalitpur, and Bhaktapur => 01
    if number.startswith("01") and code not in ["010", "011", "019"]:
        return "01"

    return code


def _parse_landline_number(number) -> dict:
    """
    Parse and returns mobile number details.
    :return:
    {
        "type": "Landline",
        "number": "98XXXXXXXX",
        "operator": <Operator>
    }
    """
    number = get_exact_number(number)

    # adding zero
    if number[0] != "0":
        number = f"0{number}"

    return {
        "type": "Landline",
        "number": number,
        "area_code": _get_area_code(number),
    }



import datetime as pythonDateTime

from nepali.date_converter import converter as nepali_date_converter
from nepali.timezone import NepaliTimeZone, to_nepali_timezone, utc_now

from ._nepalimonth import nepalimonth


class formatter_class_mixin:
    """
    mixin for date time formatter.
    adds methods `strftime(format)`
    """

    def get_formatter_class(self):
        return self.__class__.init_formatter_class()

    @classmethod
    def init_formatter_class(cls):
        if not hasattr(cls, "_formatter_class_cache"):
            from ._formatter import NepaliDateTimeFormatter

            cls._formatter_class_cache = NepaliDateTimeFormatter
        return cls._formatter_class_cache

    @classmethod
    def get_strptime_method(cls):
        if not hasattr(cls, "_strptime_method_cache"):
            from .parser import strptime

            cls._strptime_method_cache = strptime
        return cls._strptime_method_cache

    def strftime(self, format: str) -> str:
        return self.strftime_en(format)

    def strftime_en(self, format: str) -> str:
        nepali_datetime_formatter = self.get_formatter_class()
        formatter = nepali_datetime_formatter(self, devanagari=False)
        return formatter.get_str(format)

    def strftime_ne(self, format: str) -> str:
        nepali_datetime_formatter = self.get_formatter_class()
        formatter = nepali_datetime_formatter(self, devanagari=True)
        return formatter.get_str(format)


class nepalidate(formatter_class_mixin):
    def __init__(self, year, month, day) -> None:
        if isinstance(month, nepalimonth):
            month = month.value

        self.__year = year
        self.__month = month
        self.__day = day

        # converting to english date
        year_en, month_en, day_en = nepali_date_converter.nepali_to_english(
            year, month, day
        )
        self.__python_date = pythonDateTime.date(year_en, month_en, day_en)

    def __str__(self):
        return self.strftime_en("%Y-%m-%d")

    def __repr__(self):
        return f"<nepalidate> {self}"

    def to_datetime(self):
        return pythonDateTime.datetime.combine(
            self.__python_date, pythonDateTime.time(), tzinfo=NepaliTimeZone()
        )

    def to_date(self):
        return self.__python_date

    def to_nepalidatetime(self):
        return nepalidatetime.from_nepali_date(self)

    # operators overloading

    def __add__(self, other):
        """addition"""
        if type(other) == pythonDateTime.timedelta:
            # timedelta object
            return nepalidate.from_date(self.to_date() + other)
        return NotImplemented

    def __sub__(self, other):
        """subtraction"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() - other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() - other

        elif type(other) == pythonDateTime.timedelta:
            # timedelta object
            return nepalidate.from_date(self.to_date() - other)

        return NotImplemented

    def __lt__(self, other):
        """less than"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_date() < other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() < other

        return NotImplemented

    def __le__(self, other):
        """less than equal"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() <= other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() <= other

        return NotImplemented

    def __eq__(self, other):
        """equal"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() == other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() == other

        return False

    def __ne__(self, other):
        """not equal"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() != other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() != other

        return True

    def __gt__(self, other):
        """greater than"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() > other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() > other

        return NotImplemented

    def __ge__(self, other):
        """greater than equal"""
        if type(other) == self.__class__:
            # nepalidate object
            return self.to_date() >= other.to_date()

        elif type(other) == pythonDateTime.date:
            # python date object
            return self.to_date() >= other

        return NotImplemented

    # static methods
    @classmethod
    def strptime(cls, datetime_str, format):
        nepalidatetime_strptime = cls.get_strptime_method()
        return nepalidatetime_strptime(datetime_str, format=format).date()

    @staticmethod
    def now():
        return nepalidate.today()

    @staticmethod
    def today():
        return nepalidate.from_date(pythonDateTime.date.today())

    @staticmethod
    def from_datetime(datetime_object):
        return nepalidate.from_date(datetime_object.date())

    @staticmethod
    def from_date(date_object):
        return nepalidate(
            *nepali_date_converter.english_to_nepali(
                date_object.year, date_object.month, date_object.day
            )
        )

    @staticmethod
    def from_nepalidatetime(datetime_object):
        return datetime_object.date()

    # property

    def weekday(self):
        """
        Sunday => 0, Saturday => 6
        """
        return (self.__python_date.weekday() + 1) % 7

    # nepali date properties
    @property
    def year(self):
        return self.__year

    @property
    def month(self):
        return self.__month

    @property
    def day(self):
        return self.__day


class nepalitime(pythonDateTime.time):
    def __repr__(self):
        return f"<nepalitime> {self}"

    # static methods

    @staticmethod
    def now():
        dt_now = pythonDateTime.datetime.now()
        return nepalitime(dt_now.hour, dt_now.minute, dt_now.second, dt_now.microsecond)


class nepalidatetime(formatter_class_mixin):
    """
    nepalidatetime
    """

    def __init__(self, year, month, day, hour=0, minute=0, second=0, microsecond=0):
        self.__np_date = nepalidate(year, month, day)
        self.__np_time = nepalitime(hour, minute, second, microsecond)

    def __str__(self):
        return f"{self.__np_date} {self.__np_time}"

    def __repr__(self):
        return f"<nepalidatetime> {self}"

    # operator overloading

    def __add__(self, other):
        """addition"""
        if type(other) == pythonDateTime.timedelta:
            # timedelta object
            return nepalidatetime.from_datetime(self.to_datetime() + other)

        return NotImplemented

    def __sub__(self, other):
        """subtraction"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() - other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() - to_nepali_timezone(other)

        elif type(other) == pythonDateTime.timedelta:
            # timedelta object
            return nepalidatetime.from_datetime(self.to_datetime() - other)

        return NotImplemented

    def __lt__(self, other):
        """less than"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() < other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() < to_nepali_timezone(other)

        return NotImplemented

    def __le__(self, other):
        """less than equal"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() <= other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() <= to_nepali_timezone(other)

        return NotImplemented

    def __eq__(self, other):
        """equal"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() == other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() == to_nepali_timezone(other)

        return False

    def __ne__(self, other):
        """not equal"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() != other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() != to_nepali_timezone(other)

        return True

    def __gt__(self, other):
        """greater than"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() > other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() > to_nepali_timezone(other)

        return NotImplemented

    def __ge__(self, other):
        """greater than equal"""
        if type(other) == self.__class__:
            # nepalidatetime object
            return self.to_datetime() >= other.to_datetime()

        elif type(other) == pythonDateTime.datetime:
            # python datetime object
            return self.to_datetime() >= to_nepali_timezone(other)

        return NotImplemented

    # object transformation

    def to_datetime(self):
        return pythonDateTime.datetime.combine(
            self.__np_date.to_date(), self.__np_time, tzinfo=NepaliTimeZone()
        )

    def to_date(self):
        return self.to_datetime().date()

    def date(self):
        return self.__np_date

    def time(self):
        return self.__np_time

    # static methods

    @classmethod
    def strptime(cls, datetime_str, format):
        nepalidatetime_strptime = cls.get_strptime_method()
        return nepalidatetime_strptime(datetime_str, format=format)

    @staticmethod
    def from_datetime(dt):
        dt = to_nepali_timezone(dt)
        nd = nepalidate.from_date(dt.date())
        return nepalidatetime(
            nd.year, nd.month, nd.day, dt.hour, dt.minute, dt.second, dt.microsecond
        )

    @staticmethod
    def from_date(date_object):
        nepali_date_object = nepalidate.from_date(date_object)
        return nepalidatetime.from_nepali_date(nepali_date_object)

    @staticmethod
    def from_nepali_date(nepali_date_object):
        return nepalidatetime(
            nepali_date_object.year, nepali_date_object.month, nepali_date_object.day
        )

    @staticmethod
    def now():
        return nepalidatetime.from_datetime(utc_now())

    # property

    @property
    def year(self):
        return self.__np_date.year

    @property
    def month(self):
        return self.__np_date.month

    @property
    def day(self):
        return self.__np_date.day

    def weekday(self):
        """
        Sunday => 0, Saturday => 6
        """
        return self.__np_date.weekday()

    @property
    def hour(self):
        return self.__np_time.hour

    @property
    def minute(self):
        return self.__np_time.minute

    @property
    def second(self):
        return self.__np_time.second
